"""Positive messages to encourage users in hard times."""
import collections

Quote = collections.namedtuple('Quote', ('content', 'author'))

# https://liveboldandbloom.com/07/quotes/words-of-encouragement
QUOTES = [
    Quote("The only person you are destined to become is the person you decide to be.", "Ralph Waldo Emerson"),
    Quote("Start where you are. Use what you have. Do what you can.", "Arthur Ashe"),
    Quote("How wonderful it is that nobody need wait a single moment before starting to improve the world.", "Anne Frank"),
    Quote("When one door closes another door opens; but we so often look so long and so regretfully upon the closed door, that we do not see the ones which open for us.", "Alexander Graham Bell"),
    Quote("Life is a succession of lessons which must be lived to be understood.", "Helen Keller"),
    Quote("When you get into a tight place and everything goes against you, till it seems as though you could not hang on a minute longer, never give up then, for that is just the place and time that the tide will turn.", "Harriet Beecher Stowe"),
    Quote("You must do the thing you think you cannot do.", "Eleanor Roosevelt"),
    Quote("If you don’t pay appropriate attention to what has your attention, it will take more of your attention than it deserves.", "David Allen"),
    Quote("I find hope in the darkest of days, and focus in the brightest. I do not judge the universe.", "Dalai Lama"),
    Quote("Character cannot be developed in ease and quiet. Only through experience of trial and suffering can the soul be strengthened, ambition inspired, and success achieved.", "Helen Keller"),
    Quote("It is by going down into the abyss that we recover the treasures of life. Where you stumble, there lies your treasure.", "Joseph Campbell"),
    Quote("In essence, if we want to direct our lives, we must take control of our consistent actions. It’s not what we do once in a while that shapes our lives, but what we do consistently.", "Tony Robbins"),
    Quote("Our greatest weakness lies in giving up. The most certain way to succeed is always to try just one more time.", "Thomas A. Edison"),
    Quote("You are never to old to set another goal or to dream a new dream.", "C.S. Lewis"),
    Quote("Even if you fall on your face, you're still moving forward.", "Victor Kiam"),
    Quote("Be miserable. Or motivate yourself. Whatever has to be done, it's always your choice.", "Wayne Dyer"),
    Quote("Learn from the past, set vivid, detailed goals for the future, and live in the only moment of time over which you have any control: now.", "Denis Waitley"),
    Quote("Do you want to know who you are? Don't ask. Act! Action will delineate and define you.", "Thomas Jefferson"),
    Quote("The key is to keep company only with people who uplift you, whose presence calls forth your best.", "Epictetus"),
    Quote("Be impeccable with your word. Speak with integrity. Say only what you mean. Avoid using the word to speak against yourself or to gossip about others. Use the power of your word in the direction of truth and love.", "Miguel Angel Ruiz"),
    Quote("Act as if what you do makes a difference. It does.", "William James"),
    Quote("Learning is the beginning of wealth. Learning is the beginning of health. Learning is the beginning of spirituality. Searching and learning is where the miracle process all begins.", "Jim Rohn"),
    Quote("I'd rather attempt to do something great and fail than to attempt to do nothing and succeed.", "Robert H. Schuller"),
    Quote("If it is bread that you seek, you will have bread. If it is the soul you seek, you will find the soul. If you understand this secret, you know you are that which you seek.", "Rumi"),
    Quote("Always say ‘yes' to the present moment. What could be more futile, more insane, than to create inner resistance to what already is? what could be more insane than to oppose life itself, which is now and always now? Surrender to what is. Say ‘yes' to life — and see how life suddenly starts working for you rather than against you.", "Eckhart Tolle"),
    Quote("Perhaps everything terrible is in its being something that needs our love.", "Rainer Maria Rilke"),
    Quote("Hold on to what is good, even if it's a handful of earth.", "Hopi prayer"),
    Quote("Let your soul stand cool and composed before a million universes.", "Walt Whitman"),
    Quote("If you are going through hell, keep going.", "Winston Churchill"),
    Quote("Tell you heart that the fear of suffering is worse than suffering itself. And no heart has ever suffered when it goes in search of its dream.", "Paulo Coelho"),
    Quote("The secret of health for both mind and body is not to mourn for the past, not to worry about the future . . . but to live the present moment wisely and earnestly.", "Buddha"),
    Quote("In the midst of movement and chaos, keep stillness inside of you.", "Deepak Chopra"),
    Quote("Good humor is a tonic for mind and body. It is the best antidote for anxiety and depression. It is a business asset. It attracts and keep friends. It lightens human burdens. It is the direct route to serenity and contentment.", "Greenville Kleisser"),
    Quote("What you thought before has led to every choice you have made, and this adds up to you at this moment. If you want to change who you are physically, mentally, and spiritually, you will have to change what you think.", "Dr. Patrick Gentempo"),
    Quote("Too often we underestimate the power of a touch, a smile, a kind word, a listening ear, an honest compliment, or the smallest act of caring, all of which have the potential to turn a life around.", "Leo Buscaglia"),
    Quote("Fall seven times, stand up eight.", "Japanese Proverb"),
    Quote("Problems are not stop signs, they are guidelines.", "Robert Schuller"),
    Quote("We must embrace pain and burn it as fuel for our journey.", "Kenji Miyazawa"),
    Quote("By recording your dreams and goals on paper, you set in motion the process of becoming the person you most want to be. Put your future in good hands — your own.", "Mark Victor Hansen"),
    Quote("Be patient with yourself. Self-growth is tender; it’s holy ground. There’s no greater investment.", "Stephen Covey"),
    Quote("Would you like me to give you a formula for success? It’s quite simple, really. Double your rate of failure. You are thinking of failure as the enemy of success. But it isn’t at all. You can be discouraged by failure or you can learn from it, So go ahead and make mistakes. Make all you can. Because remember that’s where you will find success.", "Thomas J. Watson"),
    Quote("Nobody can go back and start a new beginning, but anyone can start today and make a new ending.", "Maria Robinson"),
    Quote("If we did all the things we are capable of, we would astound ourselves.", "Thomas Edison"),
    Quote("Try not to become a person of success, but rather try to become a person of value.", "Albert Einstein"),
    Quote("If you don't value your time, neither will others. Stop giving away your time and talents — start charging for it.", "Kim Garst"),
    Quote("Don't be afraid to give up the good to go for the great.", "John D. Rockefeller"),
    Quote("Life is not about finding yourself. Life is about creating yourself.", "Lolly Daskal"),
    Quote("All progress takes place outside the comfort zone.", "Michael John Bobak"),
    Quote("When I dare to be powerful – to use my strength in the service of my vision, then it becomes less and less important whether I am afraid.", "Audre Lorde"),
    Quote("A man can be as great as he wants to be. If you believe in yourself and have the courage, the determination, the dedication, the competitive drive and if you are willing to sacrifice the little things in life and pay the price for the things that are worthwhile, it can be done.", "Vince Lombardi"),
]

__all__ = ['QUOTES']



"""
Java Backend Format Strings

The following section lists format strings for the Java backend, and can/
should be updated as the text API to the Stanford Portable Library changes.
The strings are marked using keyword format syntax, which can be expanded by
keyword in Python
"""
# Platform::filelib_fileExists
# Platform::filelib_isFile
# Platform::filelib_isSymbolicLink
# Platform::filelib_isDirectory
# Platform::filelib_setCurrentDirectory
# Platform::filelib_getCurrentDirectory
# Platform::filelib_getTempDirectory
# Platform::filelib_createDirectory
# Platform::filelib_deleteFile
# Platform::filelib_getDirectoryPathSeparator
# Platform::filelib_getSearchPathSeparator
# Platform::filelib_expandPathname
# Platform::filelib_listDirectory
# Platform::filelib_fileExists
# Platform::filelib_isFile
# Platform::filelib_isSymbolicLink
# Platform::filelib_isDirectory
# Platform::filelib_setCurrentDirectory
# Platform::filelib_getCurrentDirectory
# Platform::filelib_getTempDirectory
# Platform::filelib_createDirectory
# Platform::filelib_deleteFile
# Platform::filelib_getDirectoryPathSeparator
# Platform::filelib_getSearchPathSeparator
# Platform::filelib_expandPathname
# Platform::filelib_listDirectory

# Platform::setStackSize

# Platform::os_getLastError

# Platform::regex_match
# Platform::regex_matchCount
# Platform::regex_matchCountWithLines
# Platform::regex_replace

# Transfer to functions: DONE:
# File_


# TODO: special
File_openFileDialog = 'File.openFileDialog("{title!q}", "{mode}", "{path!q}")'   # TODO check if needs to be quoted

GWindow_constructor = 'GWindow.create("{id}", {width}, {height}, "{top_compound}", {visible!b})'
GWindow_delete = 'GWindow.delete("{id}")'
GWindow_close = 'GWindow.close("{id}")'
GWindow_requestFocus = 'GWindow.requestFocus("{id}")'
GWindow_setExitOnClose = 'GWindow.setExitOnClose("{id}", {value!b})'
GWindow_clear = 'GWindow.clear("{id}")'
GWindow_clearCanvas = 'GWindow.clearCanvas("{id}")'
GWindow_repaint = 'GWindow.repaint("{id}")'
GWindow_saveCanvasPixels = 'GWindow.saveCanvasPixels("{id}", {filename!q}'
GWindow_setSize = 'GWindow.setSize("{id}", {width}, {height})'
GWindow_setCanvasSize = 'GWindow.setCanvasSize("{id}", {width}, {height})'
GWindow_setCloseOperation = 'GWindow.setCloseOperation("{id}", {op}'
GWindow_minimize = 'GWindow.minimize("{id}")'
GWindow_pack = 'GWindow.pack("{id}")'
GWindow_setTitle = 'GWindow.setTitle("{id}", {title!q})'
GWindow_setLocation = 'GWindow.setLocation("{id}", {x}, {y})'
GWindow_setLocationSaved = 'GWindow.setLocationSaved("{id}", {value!b})'
GWindow_setPixel = 'GWindow.setPixel("{id}", {x}, {y}, {rgb}, {repaint!b})'
GWindow_setPixels = 'GWindow.setPixels("{id}", {base64!q})' # TODO: Special
GWindow_toBack = 'GWindow.toBack("{id}")'
GWindow_toFront = 'GWindow.toFront("{id}")'
GWindow_getScreenHeight = 'GWindow.getScreenHeight()'
GWindow_getScreenSize = 'GWindow.getScreenSize()'
GWindow_getScreenWidth = 'GWindow.getScreenWidth()'
GWindow_draw = 'GWindow.draw("{id}", "{gobj_id}")'
GWindow_drawInBackground = 'GWindow.drawInBackground("{id}", "{gobj_id}")'
GWindow_exitGraphics = 'GWindow.exitGraphics()' # TODO: special
GWindow_setRegionAlignment = 'GWindow.setRegionAlignment("{id}", "{region}", "{align}")'
GWindow_setResizable = 'GWindow.setResizable("{id}", {value!b})'
GWindow_addToRegion = 'GWindow.addToRegion("{id}", "{gobj_id}", "{region}")'
GWindow_getLocation = 'GWindow.getLocation("{id}")' # TODO special
GWindow_getPixel = 'GWindow.getPixel("{id}", {x}, {y})' # TODO special
GWindow_getPixels = 'GWindow.getPixels("{id}")'  # TODO special
GWindow_getRegionSize = 'GWindow.getRegionSize("{id}", "{region}")' # TODO: special
GWindow_removeFromRegion = 'GWindow.removeFromRegion("{id}", "{gobj_id}", "{region}")'
GWindow_getSize = 'GWindow.getSize("{id}")' # TODO special
GWindow_getCanvasSize = 'GWindow.getCanvasSize("{id}")'  # TODO: special
GWindow_getContentPaneSize = 'GWindow.getContentPaneSize("{id}")' # TODO: special
GWindow_setVisible = 'GWindow.setVisible("{id}", {flag!b})'

GTimer_constructor = 'GTimer.create("{id}", {millis})'
GTimer_delete = 'GTimer.deleteTimer("{id}")'
GTimer_start = 'GTimer.startTimer("{id}")'
GTimer_pause = 'GTimer.pause({millis})'
GTimer_stop = 'GTimer.stopTimer("{id}")'

HttpServer_sendResponse = 'HttpServer.sendResponse({request_id}, {http_error_code}, {contentType!q}, {responseText!q})'
HttpServer_sendResponseFile = 'HttpServer.sendResponseFile({request_id}, {content_type!q}, {response_file_path!q})'
HttpServer_start = 'HttpServer.start({port})'
HttpServer_stop = 'HttpServer.stop()'

Sound_constructor = 'Sound.create("{id}", {filename!q})'
Sound_delete = 'Sound.delete("{id}")'
Sound_play = 'Sound.play("{id}")'

# Platform::url_download

GCompound_constructor = 'GCompound.create("{id}")'
GCompound_add = 'GCompound.add("{compound_id}", "{gobj_id}")'

GObject_delete = 'GObject.delete("{id}")'
GObject_remove = 'GObject.remove("{id}")'
GObject_sendForward = 'GObject.sendForward("{id}")'
GObject_sendToFront = 'GObject.sendToFront("{id}")'
GObject_sendBackward = 'GObject.sendBackward("{id}")'
GObject_sendToBack = 'GObject.sendToBack("{id}")'
GObject_setVisible = 'GObject.setVisible("{id}", {flag!b})'
GObject_setColor = 'GObject.setColor("{id}", "{color}")'
GObject_scale = 'GObject.scale("{id}", {sx}, {sy})'
GObject_rotate = 'GObject.rotate("{id}", {theta})'
GObject_contains = 'GObject.contains("{id}", {x}, {y})'  # TODO: special move into gobjects.py
GObject_getBounds = 'GObject.getBounds("{id}")' # TODO: special
GObject_setLineWidth = 'GObject.setLineWidth("{id}", {line_width})'
GObject_setLocation = 'GObject.setLocation("{id}", {x}, {y})'
GObject_setSize = 'GObject.setSize("{id}", {width}, {height})'
GObject_setAntialiasing = 'GObject.setAntialiasing({value!b})'
GObject_setFilled = 'GObject.setFilled("{id}", {flag!b})'
GObject_setFillColor = 'GObject.setFillColor("{id}", "{color}")'

GInteractor_isEnabled = 'GInteractor.isEnabled("{id}")' # TODO: special
GInteractor_setEnabled = 'GInteractor.setEnabled("{id}", {value!b})'
GInteractor_setFont = 'GInteractor.setFont("{id}", "{font!u}")'
GInteractor_setIcon = 'GInteractor.setIcon("{id}", {filename!q})'
GInteractor_setMnemonic = 'GInteractor.setMnemonic("{id}", {mnemonic})' # TODO: special
GInteractor_setText = 'GInteractor.setText("{id}", {text!u})'
GInteractor_setTextPosition = 'GInteractor.setTextPosition("{id}", {horizontal}, {vertical})'
GInteractor_setTooltip = 'GInteractor.setTooltip("{id}", {tooltip_text!q})'

GRect_constructor = 'GRect.create("{id}", {width}, {height})'

GRoundRect_constructor = 'GRoundRect.create("{id}", {width}, {height}, {corner})'

G3DRect_constructor = 'G3DRect.create("{id}", {width}, {height}, {raised!b})'
G3DRect_setRaised = 'G3DRect.setRaised("{id}", {raised!b})'

GOval_constructor = 'GOval.create("{id}", {width}, {height})'

GArc_constructor = 'GArc.create("{id}", {width}, {height}, {start}, {sweep})'
GArc_setStartAngle = 'GArc.setStartAngle("{id}", {angle})'
GArc_setSweepAngle = 'GArc.setSweepAngle("{id}", {angle})'
GArc_setFrameRectangle = 'GArc.setFrameRectangle("{id}", {x}, {y}, {width}, {height})' # TODO: special

GLine_constructor = 'GLine.create("{id}", {x1}, {y1}, {x2}, {y2})'
GLine_setStartPoint = 'GLine.setStartPoint("{id}", {x}, {y})' # TODO special
GLine_setEndPoint = 'GLine.setEndPoint("{id}", {x}, {y})'

GImage_constructor = 'GImage.create("{id}", "{filename}")' # TODO: special

GLabel_constructor = 'GLabel.create("{id}", {label!q})'
GLabel_setFont = 'GLabel.setFont("{id}", "{font}")'
GLabel_setLabel = 'GLabel.setLabel("{id}", {label!q})'
GLabel_getFontAscent = 'GLabel.getFontAscent("{id}")' # TODO: special
GLabel_getFontDescent = 'GLabel.getFontDescent("{id}")'  # TODO: special
GLabel_getSize = 'GLabel.getGLabelSize("{id}")'  # TODO: special

GPolygon_constructor = 'GPolygon.create("{id}")'
GPolygon_addVertex = 'GPolygon.addVertex("{id}", {x}, {y})'  # TODO: special

DiffImage_compareImages = 'DiffImage.compareImages({file1!q}, {file2!q}, {outfile!q})'
DiffImage_compareWindowToImage = 'DiffImage.compareWindowToImage("{id}", {file2!q}, {ignore_winodw_size!b})'
DiffImage_show = 'DiffImage.show({file1!q}, {file2!q})'

GBufferedImage_constructor = 'GBufferedImage.create("{id}", {x}, {y}, {width}, {height}, {rgb})' # TODO special and round
GBufferedImage_fill = 'GBufferedImage.fill("{id}", {rgb})'
GBufferedImage_fillRegion = 'GBufferedImage.fillRegion("{id}", {x}, {y}, {width}, {height}, {rgb})'  # TODO rounding
GBufferedImage_load = 'GBufferedImage.load("{id}", {filename!q})'
GBufferedImage_resize = 'GBufferedImage.resize("{id}", {width}, {height}, {retain!b})'  # TODO: round
GBufferedImage_save = 'GBufferedImage.save("{id}", {filename!q})'
GBufferedImage_setRGB = 'GBufferedImage.setRGB("{id}", {x}, {y}, {rgb})' # TODO round
GBufferedImage_updateAllPixels = 'GBufferedImage.updateAllPixels("{id}", "{base64}")'

# SECTION: GInteractor

GInteractor_setAccelerator = 'GInteractor.setAccelerator("{id}", {accelerator!u})'
GInteractor_setActionCommand = 'GInteractor.setActionCommand("{id}", {cmd!q})'
GInteractor_setBackground = 'GInteractor.setBackground("{id}", "{color}")'
GInteractor_addActionListener = 'GInteractor.addActionListener("{id}")'
GInteractor_removeActionListener = 'GInteractor.removeActionListener("{id}")'
GInteractor_requestFocus = 'GInteractor.requestFocus("{id}")'
GInteractor_getFont = 'GInteractor.getFont("{id}")'  # TODO: special
GInteractor_getMnemonic = 'GInteractor.getMnemonic("{id}")' # TODO: special
GInteractor_getSize = 'GInteractor.getSize("{id}")'  # TODO: special

GButton_constructor = 'GButton.create("{id}", {label!q})'  # TODO: special

GCheckBox_constructor = 'GCheckBox.create("{id}", {label!q})'
GCheckBox_isSelected = 'GCheckBox.isSelected("{id}")'  # TODO: special
GCheckBox_setSelected = 'GCheckBox.setSelected("{id}", {state!b})'

GRadioButton_constructor = 'GRadioButton.create("{id}", {label!q}, {group!q})'
GRadioButton_isSelected = 'GRadioButton.isSelected("{id}")'  # TODO: special
GRadioButton_setSelected = 'GRadioButton.setSelected("{id}", {state!b})'

GSlider_constructor = 'GSlider.create("{id}", {min}, {max}, {value})'
GSlider_getMajorTickSpacing = 'GSlider.getMajorTickSpacing("{id}")'  # TODO: special
GSlider_getMinorTickSpacing = 'GSlider.getMinorTickSpacing("{id}")'  # TODO: special
GSlider_getPaintLabels = 'GSlider.getPaintLabels("{id}")'  # TODO: special
GSlider_getPaintTicks = 'GSlider.getPaintTicks("{id}")'  # TODO: special
GSlider_getSnapToTicks = 'GSlider.getSnapToTicks("{id}")'  # TODO: special
GSlider_getValue = 'GSlider.getValue("{id}")'  # TODO: special
GSlider_setMajorTickSpacing = 'GSlider.setMajorTickSpacing("{id}", {value})'
GSlider_setMinorTickSpacing = 'GSlider.setMinorTickSpacing("{id}", {value})'
GSlider_setPaintLabels = 'GSlider.setPaintLabels("{id}", {value!b})'
GSlider_setPaintTicks = 'GSlider.setPaintTicks("{id}", {value!b})'
GSlider_setSnapToTicks = 'GSlider.setSnapToTicks("{id}", {value!b})'
GSlider_setValue = 'GSlider.setValue("{id}", {value})'

GTable_constructor = 'GTable.create("{id}", {num_rows}, {num_cols}, {x}, {y}, {width}, {height})'  # TODO: special
GTable_autofitColumnWidths = 'GTable.autofitColumnWidths("{id}")'
GTable_clear = 'GTable.clear("{id}")'
GTable_clearFormatting = 'GTable.clearFormatting("{id}")'
GTable_get = 'GTable.get("{id}", {row}, {column})'  # TODO: special
GTable_getColumnWidth = 'GTable.getColumnWidth("{id}", {column})'  # TODO: special
GTable_getSelection = 'GTable.getSelection("{id}")'  # TODO: special
GTable_resize = 'GTable.resize("{id}", {num_rows}, {num_cols})'  # TODO: special
GTable_select = 'GTable.select("{id}", {row}, {column})'
GTable_set = 'GTable.set("{id}", {row}, {column}, {value!q})'
GTable_setCellAlignment = 'GTable.setCellAlignment("{id}", {row}, {column}, {align})'
GTable_setCellBackground = 'GTable.setCellBackground("{id}", {row}, {column}, {color!q})'
GTable_setCellFont = 'GTable.setCellFont("{id}", {row}, {column}, {font!q})'
GTable_setCellForeground = 'GTable.setCellForeground("{id}", {row}, {column}, {color!q}))'
GTable_setColumnAlignment = 'GTable.setColumnAlignment("{id}", {column}, {align}))'
GTable_setColumnBackground = 'GTable.setColumnBackground("{id}", {column}, {color!q})'
GTable_setColumnFont = 'GTable.setColumnFont("{id}", {column}, {font!q})'
GTable_setColumnForeground = 'GTable.setColumnForeground("{id}", {column}, {color!q})'
GTable_setColumnHeaderStyle = 'GTable.setColumnHeaderStyle("{id}", {style})'
GTable_setColumnWidth = 'GTable.setColumnWidth("{id}", {column}, {width})'
GTable_setEditable = 'GTable.setEditable("{id}", {editable!b})'
GTable_setEditorValue = 'GTable.setEditorValue("{id}", {row}, {column}, {value!q})'
GTable_setEventEnabled = 'GTable.setEventEnabled("{id}", {type}, {enabled!b})'
GTable_setFont = 'GTable.setFont("{id}", {font!q})'
GTable_setHorizontalAlignment = 'GTable.setHorizontalAlignment("{id}", {alignment!q})'
GTable_setRowAlignment = 'GTable.setRowAlignment("{id}", {row}, {align})'
GTable_setRowBackground = 'GTable.setRowBackground("{id}", {row}, {color!q})'
GTable_setRowColumnHeadersVisible = 'GTable.setRowColumnHeadersVisible("{id}", {visible!b})'
GTable_setRowFont = 'GTable.setRowFont("{id}", {row}, {font!q})'
GTable_setRowForeground = 'GTable.setRowForeground("{id}", {row}, {color!q})'

GTextArea_constructor = 'GTextArea.create("{id}", {width}, {height})'
GTextArea_getText = 'GTextArea.getText("{id}")'  # TODO: special
GTextArea_setEditable = 'GTextArea.setEditable("{id}", {editable!b})'
GTextArea_setFont = 'GTextArea.setFont("{id}", {font!q})'
GTextArea_setText = 'GTextArea.setText("{id}", {text!q})'

GTextField_constructor = 'GTextField.create("{id}", {num_chars})'  # TODO: special
GTextField_getText = 'GTextField.getText("{id}")'  # TODO: special
GTextField_isEditable = 'GTextField.isEditable("{id}")'  # TODO: special
GTextField_setEditable = 'GTextField.setEditable("{id}", {editable!b})'
GTextField_setPlaceholder = 'GTextField.setPlaceholder("{id}", {text!q})'
GTextField_setText = 'GTextField.setText("{id}", {text!q})'

GChooser_constructor = 'GChooser.create("{id}")'  #TODO: special
GChooser_addItem = 'GChooser.addItem("{id}", {item!q})'
GChooser_getSelectedItem = 'GChooser.getSelectedItem("{id}")' # TODO: special
GChooser_setSelectedItem = 'GChooser.setSelectedItem("{id}", {item!q})'

# END SECTION: GInteractor

GEvent_getNextEvent = 'GEvent.getNextEvent({mask})' # TODO: special
GEvent_waitForEvent = 'GEvent.waitForEvent({mask})' # TODO: special

GFileChooser_showOpenDialog = 'GFileChooser.showOpenDialog({current_dir!q}, {file_filter!q})'  # TODO: special
GFileChooser_showSaveDialog = 'GFileChooser.showSaveDialog({current_dir!q}, {file_filter!q})' # TODO: special

GOptionPane_showConfirmDialog = 'GOptionPane.showConfirmDialog({message!u}, {title!u}, {type})' # TODO: special
GOptionPane_showInputDialog = 'GOptionPane.showInputDialog({message!u}, {title!u})'  # TODO: special
GOptionPane_showMessageDialog = 'GOptionPane.showMessageDialog({message!u}, {title!u}, {type})'  # TODO: special
GOptionPane_showOptionDialog = 'GOptionPane.showOptionDialog({message!u}, {title!u}, {options}, {initially_selected!q})'  #TODO: very special
GOptionPane_showTextFileDialog = 'GOptionPane.showTextFileDialog({message!u}, {title!u}, {rows}, {cols})' # TODO: special

Clipboard_get = 'Clipboard.get()'  # TODO: special
Clipboard_set = 'Clipboard.set({text!u})'

# Platform::cpplib_setCppLibraryVersion
# Platform::cpplib_getCppLibraryVersion
SPL_getJavaBackEndVersion = 'StanfordCppLib.getJbeVersion()' # TODO: special

# JBEConsole_isBlocked = 'JBEConsole.isBlocked()'
JBEConsole_print = 'JBEConsole.print({line!q}, {stderr!b})'  # TODO: special
JBEConsole_getLine = 'JBEConsole.getLine()'  # TODO: special
JBEConsole_clear = 'JBEConsole.clear()'
JBEConsole_minimize = 'JBEConsole.minimize()'
JBEConsole_setFont = 'JBEConsole.setFont({font})'
JBEConsole_setSize = 'JBEConsole.setSize({width}, {height})'
JBEConsole_setTitle = 'JBEConsole.setTitle({title!q})'
JBEConsole_setLocation = 'JBEConsole.setLocation({x}, {y})'
JBEConsole_setCloseOperation = 'JBEConsole.setCloseOperation({value})'
JBEConsole_setErrorColor = 'JBEConsole.setErrorColor({color!q})'
JBEConsole_setExitProgramOnClose = 'JBEConsole.setExitProgramOnClose({value!b})'
JBEConsole_setLocationSaved = 'JBEConsole.setLocationSaved({value!b})'
JBEConsole_setOutputColor = 'JBEConsole.setOutputColor({color!q})'
JBEConsole_setVisible = 'JBEConsole.setVisible({value!b})'
JBEConsole_toFront = 'JBEConsole.toFront()'

Note_play = 'Note.play({note!u})'  # TODO: special

AutograderInput_addButton = 'AutograderInput.addButton({text!u}, {input!u})'  # TODO: special
AutograderInput_removeButton = 'AutograderInput.removeButton({text!u})'
AutograderInput_addCategory = 'AutograderInput.addCategory({name!u})'
AutograderInput_removeCategory = 'AutograderInput.removeCategory({name!u})'
AutograderInput_setVisible = 'AutograderInput.setVisible({visible!b})'
AutograderUnitTest_addTest = 'AutograderUnitTest.addTest({test_name!u}, {category!u}, {style_check!b})'
AutograderUnitTest_catchExceptions = 'AutograderUnitTest.catchExceptions()' # TODO: special
AutograderUnitTest_clearTests = 'AutograderUnitTest.clearTests({style_check!b})'
AutograderUnitTest_clearTestResults = 'AutograderUnitTest.clearTestResults({style_check!b})'
AutograderUnitTest_isChecked = 'AutograderUnitTest.isChecked({full_test_name!u})' # TODO special
AutograderUnitTest_setChecked = 'AutograderUnitTest.setChecked({full_test_name!u}, {checked!b})'
AutograderUnitTest_setTestCounts = 'AutograderUnitTest.setTestCounts({pass_count}, {test_count}, {style_check!b})'
AutograderUnitTest_setTestDetails = 'AutograderUnitTest.setTestDetails({test_full_name!u}, {deets!u}, {style_check!b})'
AutograderUnitTest_setTestingCompleted = 'AutograderUnitTest.setTestingCompleted({completed!b}, {style_check!b})'
AutograderUnitTest_setTestResult = 'AutograderUnitTest.setTestResult({test_full_name!u}, {result!u}, {style_check!b})'
AutograderUnitTest_setTestRuntime = 'AutograderUnitTest.setTestRuntime({test_full_name}, {runtime_ms})'
AutograderUnitTest_setVisible = 'AutograderUnitTest.setVisible({visible!b}, {style_check!b})'
AutograderUnitTest_setWindowDescriptionText = 'AutograderUnitTest.setWindowDescriptionText({text!u}, {style_check!b})'



#!/usr/bin/env python3 -tt
"""
File: main.py
-------------
Exports a decorator to wrap a student's main function, providing
better error handling in some cases.
"""
from functools import wraps as _wraps

class DuplicateMainException(Exception):
    """Custom exception for duplicate entry points."""
    pass

def entrypoint(main):
    if entrypoint._EXISTS:
        raise DuplicateMainException("You can have at most one entry point in your code!")
    entrypoint._EXISTS = True
    @_wraps(main)
    def wrapper():
        try:
            main()
        except Exception as e:
            print("Error: {}".format(e))
            raise
    return wrapper

entrypoint._EXISTS = False



"""
This file exports a :class:`Lexicon` class, a compact structure for storing a list of words.

This :class:`Lexicon` implementation is backed by a data structure called a prefix tree or trie ("try").
"""

# from ..decorators import print_args
import collections as _collections
import collections.abc as _collections_abc

class Lexicon(_collections.abc.MutableSet):
    """Representation of a :class:`Lexicon`, or word list.

    The main difference between a lexicon and a dictionary is that
    a lexicon does not provide any mechanism for storing definitions;
    the lexicon contains only words, with no associated information.

    It is therefore similar to a set of strings, but with a more
    space-efficient internal representation. The :class:`Lexicon`
    class supports efficient lookup operations for words and prefixes.

    For example, the following program lists all of the two-letter words
    in the lexicon stored at `english.lex`::

        lex = Lexicon('english.lex')
        for word in lex:
            if len(word) == 2:
                print(word)

    """
    def __init__(self, file=None):
        """Initialize a new lexicon.

        The default constructor creates an empty lexicon. The second form reads
        in the contents of the lexicon from a specified data filename.

        Usage::

            lex = Lexicon()
            lex_with_words = Lexicon('english.lex')

     * Initializes a new lexicon.  The default constructor creates an empty
     * lexicon.  The second form reads in the contents of the lexicon from
     * the specified data file.  The data file must be in one of two formats:
     * (1) a space-efficient precompiled binary format or (2) a text file
     * containing one word per line.  The Stanford library distribution
     * includes a binary lexicon file named <code>English.dat</code>
     * containing a list of words in English.  The standard code pattern
     * to initialize that lexicon looks like this:
     *
     *<pre>
     *    Lexicon english("English.dat");
     *</pre>
     */
        """

        self._root = None
        self.size = 0
        if file:
            self._add_words_from_file(file)

    def _add_words_from_file(self, file, delimiter='\n'):
        with open(file, 'r') as f:
            raw = f.read()

        # TODO: make this dynamically loaded?
        for line in raw.split(delimiter):
            self.add(line)

    def add(self, word):
        if not word or not word.isalpha():
            return False
        word = word.lower()
        if not self._root:
            self._root = _TrieNode()
        return self._add_helper(self._root, word, word)

    def clear(self):
        self.size = 0
        self._root = None

    def __contains__(self, word):
        if not word or not word.isalpha():
            return False
        word = word.lower()
        return self._contains_helper(self._root, word, is_prefix=False)

    def contains_prefix(self, word):
        if not word or not word.isalpha():
            return False
        word = word.lower()
        return self._contains_helper(self._root, word, is_prefix=True)

    def remove(self, word):
        if not word or not word.isalpha():
            return False
        word = word.lower()
        return self._remove_helper(self._root, word, is_prefix=False)

    discard = remove

    def remove_prefix(self, word):
        if not word or not word.isalpha():
            return False
        word = word.lower()
        return self._remove_helper(self._root, word, is_prefix=True)

    def __len__(self):
        return self.size

    def __str__(self):
        return "Lexicon(num_words={self.size})".format(self=self)

    def __lt__(self):
        pass


    # deep copy?

    def __eq__(self, other):
        pass

    #TODO: this is gross
    def __iter__(self):
        yield from self.__iter_helper__(self._root, '')

    def __iter_helper__(self, node, sofar):
        if not node:
            return
        if node.is_word:
            yield sofar
        for letter in 'abcdefghijklmnopqrstuvwxyz':
            yield from self.__iter_helper__(node.get_child(letter), sofar + letter)


    def __hash__(self):
        pass

    def _add_helper(self, node, word, original):

        if not word:  # We've reached the end, mark it as a word
            already_exists = node.is_word
            if not already_exists:
                self.size += 1
                node.is_word = True
            return already_exists

        first, *rest = word  # 'word' -> 'w', ['o', 'r', 'd']
        child = node.get_child(first)
        if not child:
            child = node.add_child(first)
        return self._add_helper(child, rest, original)

    def _contains_helper(self, node, word, is_prefix=False):
        if not node:
            # BC: No node reaches this far, so the prefix must not exist.
            return False
        if not word:
            # BC: Found nodes this far. We found it if we're looking for a prefix or we're at a word
            return is_prefix or node.is_word
        # RC: Take one step forward
        return self._contains_helper(node.get_child(word[0]), word[1:], is_prefix)

    def _remove_helper(self, node, word, ):
        if not node:
            # BC: Dead end!
            return False
        if not word:
            # BC: We've walked down the whole tree
            pass


class _TrieNode(object):
    ALPHABET_SIZE = 26

    def __init__(self):
        self.is_word = False
        self.children = [None for _ in range(_TrieNode.ALPHABET_SIZE)]
        self.num_children = 0
        self.num_descendants = 0

    def get_child(self, letter):
        """# pre: letter is between 'a' and 'z' in lowercase"""
        return self.children[ord(letter) - ord('a')]

    def add_child(self, letter):
        child = _TrieNode()
        self.children[ord(letter) - ord('a')] = child
        self.num_children += 1
        self.num_descendants += 1
        return child

    @property
    def is_leaf_node():
        return self.num_children == 0


def _scrub(string):
    return ''.join(filter(str.islower, string))



#!/usr/bin/env python3 -tt
"""
File: strlib.py
---------------
Exports a few useful string functions that are not
included by default with Python strings.
"""
import string as _string
import urllib.parse as _parse

_STRING_DELIMITERS = ",:)}]\n"

def equals_ignore_case(s1, s2):
    return s1.lower() == s2.lower()

def string_needs_quoting(string):
    """
    Checks whether the string needs quoting in order to be handled correctly.
    """

    for ch in string:
        if ch.isspace():
            break
        if ch in _STRING_DELIMITERS:
            return True
    return False

def quote_string(string, force_quotes=True):
    """
    Returns a copy of the string surrounded by double quotes, converting
    special characters to escape sequences, as necessary. If the optional
    parameter force is explicitly set to false, quotes are included
    in the output only if necessary.

    TODO make this more pythonic
    TODO: bug: doesn't print non-printing chars to JBE correctly
    """
    force_quotes |= string_needs_quoting(string)
    out = ''
    if force_quotes:
        out += '"'

    trans = str.maketrans({
        '"': '\\"',
        '\a': '\\a',
        '\b': '\\b',
        '\f': '\\f',
        '\n': '\\n',
        '\r': '\\r',
        '\t': '\\t',
        '\v': '\\v',
        '\\': '\\\\'
    })
    return '"{}"'.format(string.translate(trans))

def html_decode(encoded_html):
    return encoded_html.replace('&lt;', '<') \
                       .replace('&gt;', '>') \
                       .replace('&quot;', '"') \
                       .replace('&amp;', '&')

def html_encode(plain_html):
    return plain_html.replace('&', '&amp;') \
                     .replace('<', '&lt;') \
                     .replace('>', '&gt;') \
                     .replace('"', '&quot;')

def url_encode(string):
    # TODO make sure this is consistent with CPP lib
    return _parse.quote_plus(string, safe='~*')

def url_decode(string):
    # TODO make sure this is consistent with CPP lib
    return _parse.unquote_plus(string)


def string_is_bool(string):
    return string in (str(False), str(True))


def string_is_integer(string, radix=10):
    try:
        out = int(string, radix)
    except ValueError:
        return False
    return True

def string_is_real(string):
    try:
        out = float(string)
    except ValueError:
        return False
    return True

string_is_double = string_is_real
string_is_long = string_is_real

"""
The following methods are included only for compatibility with the Stanford CPP Libraries
They are redundant with builtin Python functionality, and should almost never be used.
"""
def bool_to_string(b):
    return 'True' if b else 'False'

def starts_with(string, prefix):
    return string.startswith(prefix)

def ends_with(string, suffix):
    return string.endswith(suffix)




"""Provide a hierarchy of graphical shapes based on the ACM Graphics model.

This module exports the superclass :class:`GObject`, as well as derived classes:

- :class:`GRect`
- :class:`GRoundRect`
- :class:`G3DRect`
- :class:`GOval`
- :class:`GArc`
- :class:`GLine`
- :class:`GImage`
- :class:`GLabel`
- :class:`GPolygon`

There is also a :class:`GCompound` class to facilitate nested object structures.

For more information on how to use each of these objects, see the comments on
each class.
"""
# TODO(sredmond): Handle lifecycles and deletion of these objects.
import campy.graphics.gcolor as _gcolor
import campy.graphics.gfont as _gfont
import campy.graphics.gmath as _gmath
import campy.graphics.gtypes as _gtypes
import campy.private.platform as _platform

from collections.abc import MutableSequence
import pathlib
import math


class GObject:
    """The common superclass of all graphical objects that can be displayed on a graphical window.

    For examples illustrating the use of the GObject class, see the descriptions
    of the individual subclasses.
    """
    def __init__(self):
        """Initialize a GObject with reasonable default values."""
        self._x = 0.0
        self._y = 0.0

        self._color = _gcolor.GColor.BLACK
        self._line_width = 1.0  # TODO(sredmond): Lots of GObjects currently ignore self._line_width
        self._visible = True

        self._transformed = False
        self._parent = None

    @property
    def x(self):
        """Get or set the x-coordinate of this :class:`GObject`."""
        return self._x

    @x.setter
    def x(self, x):
        self.location = x, self.y

    @property
    def y(self):
        """Get or set the y-coordinate of this :class:`GObject`."""
        return self._y

    @y.setter
    def y(self, y):
        self.location = self.x, y

    @property
    def location(self):
        """Get or set the location of this object as a :class:`GPoint`.

        You can supply either a GPoint or a 2-element tuple to this property.
        Accessing this property will return a GPoint which can be unpacked as a
        tuple.
        """
        return _gtypes.GPoint(self.x, self.y)

    @location.setter
    def location(self, point):
        x, y = point
        self._x = x
        self._y = y
        _platform.Platform().gobject_set_location(self, x, y)

    def move(self, dx, dy):
        """Move this object on the screen using the supplied displacements.

        :param dx: The displacement in the x-direction.
        :param dy: The displacement in the y-direction.
        """
        self.location = (self.x + dx, self.y + dy)

    @property
    def bounds(self):
        """Get the bounding box for this object.

        The bounding box of an object is the smallest rectangle that covers
        everything drawn by the figure. The coordinates of this rectangle will
        usually, but not always, match the object's location. For example, the
        location of a :class:`GLabel` gives the coordinates of the point on the
        baseline where the string begins. However, the bounds of the GLabel
        represent a rectangle that entirely covers the window area occupied by
        the GLabel.

        Subclasses must override this property (and optionally declare a setter)
        to indicate their bounding box.

        :returns: The bounding box of this GObject.
        :rtype: :class:`GRectangle`
        """
        # TODO(sredmond): Make this an abstract method?
        return _gtypes.GRectangle(-1, -1, 0, 0)

    @property
    def width(self):
        """Get the width of this :class:`GObject`.

        The width of an object is the same as the width of its bounding box.

        :returns: The width (in pixels) of this GObject.
        """
        return self.bounds.width

    @property
    def height(self):
        """Get the height of this :class:`GObject`.

        The height of an object is the same as the height of its bounding box.

        :returns: The height (in pixels) of this GObject.
        """
        return self.bounds.height

    @property
    def size(self):
        """Get the size of this :class:`GObject` as a :class:`GDimension`.

        :returns: The size of this GObject in pixels.
        :rtype: :class:`GDimension`
        """
        return _gtypes.GDimension(self.bounds.width, self.bounds.height)

    @property
    def color(self):
        """Get or set the color used to draw this object.

        The supplied color can be any object acceptable to :class:`GColor`, so
        check that class's documentation for complete details on arguments.

        Usually, the supplied color will be one of the predefined colors:

            - BLACK
            - BLUE
            - CYAN
            - DARK_GRAY
            - GRAY
            - GREEN
            - LIGHT_GRAY
            - MAGENTA
            - ORANGE
            - PINK
            - RED
            - WHITE
            - YELLOW

        TODO(sredmond): Add more detail to the possible colors with examples.

        Accessing this property will return a :class:`GColor` which can be
        used anywhere that expects a color.
        """
        # TODO(sredmond): It looks like setting the color for a gobject won't change the outside.
        return self._color

    @color.setter
    def color(self, color):
        self._color = _gcolor.GColor.normalize(color)
        _platform.Platform().gobject_set_color(self, self.color)

    @property
    def line_width(self):
        """Get or set the line width used to draw this :class:`GObject`."""
        # TODO(sredmond): Add more documentation here.
        # TODO(sredmond): Does line width need to be an integer?
        return self._line_width

    @line_width.setter
    def line_width(self, width):
        self._line_width = width
        _platform.Platform().gobject_set_line_width(self, width)

    @property
    def visible(self):
        """Get or set whether this :class:`GObject` is visible.

        A :class:`GObject` that is visible will still exist on a GWindow once added.

        Usage::

            window = GWindow()
            oval = GOval(0, 0, 41, 41)
            window.add(oval)
            print(oval.visible)  # => True
            oval.visible = False  # Hides the oval on the window.
            print(oval.visible)  # => False
        """
        return self._visible

    @visible.setter
    def visible(self, flag):
        self._visible = flag
        _platform.Platform().gwindow_set_visible(flag, gobj=self)

    def scale(self, *scales):
        """Scale this object by the given scale factor(s).

        Most clients use the one-argument form, which scales an object in both
        dimensions. To double the size of a GOval::

            oval = GOval(0, 0, 41, 41)
            oval.scale(2)
            print(oval.width, oval.height)  # => 82, 82

        There is also a two-argument form which applies separate scale factors
        to the x- and y-dimensions::

            oval = GOval(0, 0, 41, 41)
            oval.scale(3, 5)
            print(oval.width, oval.height)  # => 123, 205

        It is an error to call this function with no arguments, or with three or
        more arguments.

        :param scales: The scale factors by which to scale this object.
        """
        # TODO(sredmond): Does this scale about the center or about the origin?
        if not scales or len(scales) > 2:
            return  # TODO(sredmond): Actually fail if the number of scale factors isn't 1 or 2.

        if len(scales) == 1:
            sx, sy = scales[0], scales[0]
        else:
            sx, sy = scales

        # Mark this object as transformed so we know to defer to the platform's methods.
        self._transformed = True
        _platform.Platform().gobject_scale(self, sx, sy)

    def rotate(self, theta):
        """Rotate this object some degrees counterclockwise about its origin.

        :param: The angle (in degrees) about which to rotate this object about its origin.
        """
        # Mark this object as transformed so we know to defer to the platform's methods.
        self._transformed = True
        _platform.Platform().gobject_rotate(self, theta)

    def send_forward(self):
        """Moves this object one step toward the front in the z dimension.

        If it was already at the front of the stack, nothing happens.
        """
        if self.parent:
            self.parent.send_forward(self)

    def send_to_front(self):
        """Move this object to the front of the display in the z dimension.

        By moving it to the front, this object will appear to be on top of the
        other graphical objects on the display and may hide any objects that
        are further back.
        """
        if self.parent:
            self.parent.send_forward(self)

    def send_backward(self):
        """Moves this object one step toward the back in the z dimension.

        If it was already at the back of the stack, nothing happens.
        """
        if self.parent:
            self.parent.send_backward(self)

    def send_to_back(self):
        """Move this object to the back of the display in the z dimension.

        By moving it to the back, this object will appear to be behind the other
        graphical objects on the display and may be obscured by other objects
        that are further forward.
        """
        if self.parent:
            self.parent.send_to_back(self)

    def __contains__(self, point):
        """Implement ``point in self``.

        Check whether a given :class:`GPoint` or 2-element tuple is contained
        within the bounding box of this :class:`GObject`.

        Subclasses can override this method to provide shape-specific containment
        checks. For example, the :class:`GOval` class provides a custom method to
        check for containment.
        """
        # TODO(sredmond): It sort of feels like all of the subclasses (except GImage and GLabel)
        # just define their own contains anyway. Is it still good design to keep this implementation here?
        # Attempt to unpack the supplied point as a tuple.
        x, y = point

        # TODO(sredmond): Should we really just offload the work if it happens to be transformed?
        if self._transformed:
            return _platform.Platform().gobject_contains(self, x, y)

        return (x, y) in self.bounds

    @property
    def parent(self):
        """Return the GCompound that contains this object, if it exists.

        Each :class:`GWindow` has a single :class:`GCompound` that is aligned
        with the window. Adding objects to the window adds them to the window's
        top :class:`GCompound`. So, a :class:`GObject` does not have a parent
        until it is added to a :class:`GCompound`, either directly or by way of
        adding the :class:`GObject` to a :class:`GWindow`::

            window = GWindow()
            oval = GOval(0, 0, 41, 41)
            print(oval.parent is None)  # => True
            window.add(oval)
            print(oval.parent is None)  # => False

        :returns: This object's parent :class:`GCompound`, if it exists.
        :rtype: :class:`GCompound`
        """
        return self._parent

# SECTION: Graphical Mixins
class GFillableObject(GObject):
    """Representation of a :class:`GObject` that can be filled.

    Subclasses of :class:`GFillableObject` have a boolean property ``filled``
    and a :class:`GColor`-valued property ``fill_color``.
    """
    def __init__(self, filled=False, fill_color=None):
        super().__init__()
        self._filled = False
        self._fill_color = self.color  # Default to the outline color.

    @property
    def filled(self):
        """Get or set whether this object is filled.

        A value of True corresponds to filled and a value of False corresponds to outlined.
        """
        return self._filled

    @filled.setter
    def filled(self, is_filled):
        self._filled = is_filled
        _platform.Platform().gobject_set_filled(self, is_filled)

    @property
    def fill_color(self):
        """Return the color used to fill this object.

        If none has been set, return the empty string."""
        return self._fill_color

    @fill_color.setter
    def fill_color(self, color):
        color = _gcolor.GColor.normalize(color)
        self._fill_color = color
        _platform.Platform().gobject_set_fill_color(self, self.fill_color)

# END SECTION: Graphical Mixins

class GRect(GFillableObject):
    """Graphical representation of a rectangular box.

    To add a filled red rectangle with width 200 and height 100 to a :class:`GWindow`::

        window = GWindow()
        rect = GRect(200, 100)
        rect.filled = True
        rect.color = "RED"
        window.add(rect, 0, 0)
    """
    def __init__(self, width, height, *, x=0, y=0):
        """Create a rectangle of a width and height with an optional location.

        The two required arguments represent the width and the height of the
        :class:`GRect`, in pixels. Keyword arguments can be supplied to set the
        initial location of the :class:`GRect`.

        A :class:`GRect` defaults to being unfilled and outlined in black. The
        default location for a new :class:`GRect` is at (0, 0).

        To create a :class:`GRect` with width 200 and height 100::

            rect = GRect(200, 100)

        To create a 300x400 :class:`GRect` at position (50, 80)

            rect = GRect(200, 100, x=50, y=80)

        :param width: The width of the rectangle in pixels.
        :param height: The height of the rectangle in pixels.
        :param x: The x-coordinate of the top-left corner of this rectangle.
        :param y: The y-coordinate of the top-left corner of this rectangle.
        """
        super().__init__()
        self._width = width
        self._height = height
        self.location = x, y

    # @property
    # def size(self):
    #     return self.width, self.height

    # TODO(sredmond): Figure out how to set width as a property from an inherited getter.

    # @property
    # def setSize(self, size=None, width=None, height=None):
    #     '''
    #     Changes the size of this rectangle to the specified width and height.

    #     @type size: GDimention
    #     @type width: float
    #     @type height: float
    #     @param size: GDimension, will override height and width
    #     @rtype: void
    #     '''
    #     if(size != None):
    #         width = size.getWidth()
    #         height = size.getHeight()

    #     if(width == None or height == None): return

    #     if(self._transformed):
    #         raise Exception("setSize: Object has been transformed")

    #     self.width = width
    #     self.height = height
    #     _platform.Platform().gobject_set_size(self, width, height)

    # def setBounds(self, bounds=None, x=None, y=None, width=None, height=None):
    #     '''
    #     Changes the bounds of this rectangle to the specified values.

    #     @type bounds: GRectangle
    #     @type x: float
    #     @type y: float
    #     @type width: float
    #     @type height: float
    #     @param bounds: bounding rectangle, will override other parameters
    #     @rtype: void
    #     '''
    #     if(bounds != None):
    #         x = bounds.getX()
    #         y = bounds.getY()
    #         width = bounds.getWidth()
    #         height = bounds.getHeight()

    #     if(x == None or y == None or width == None or height == None): return

    #     self.setLocation(x=x, y=y)
    #     self.setSize(width=width, height=height)

    @property
    def bounds(self):
        """Get the bounding box for this :class:`GRect`.

        The bounding box of a :class:`GRect` is simply the rectangle's bounds
        itself.

        :returns: A bounding box that covers this :class:`GRect`.
        :rtype: :class:`GRectangle`
        """
        # TODO(sredmond): What do we we do about transformed rectangles?
        if(self._transformed): return _platform.Platform().gobject_get_bounds(self)
        return _gtypes.GRectangle(self.x, self.y, self._width, self._height)

    def __str__(self):
        return "GRect({self.width}, {self.height}, x={self.x}, y={self.y})".format(self=self)


class GRoundRect(GRect):
    """Graphical representation of a rectangular box with rounded corners.

    The rounded corners are quarter-circle arcs of a fixed diameter.
    """
    # TODO(sredmond): Add documentation from the GRect object here

    # The number of pixels in the diameter of the arc forming the corner.
    CORNER_ROUNDING = 10

    def __init__(self, width, height, *, x=None, y=None, corner=CORNER_ROUNDING):
        """Create a new rounded rectangle with the supplied width and height.

        The caller can also specify x and y coordinates for the top-left corner
        of this :class:`GRoundRect`'s bounding box.

        The corner parameter gives the diameter of the arc forming the corner,
        with a reasonable default. The higher this number is, the more rounded
        the resulting rectangle will be.

        :param width: The width of the rounded rectangle in pixels.
        :param height: The height of the rounded rectangle in pixels.
        :param x: The x-coordinate of the top-left corner of this rounded rectangle.
        :param y: The y-coordinate of the top-left corner of this rounded rectangle.
        :param corner: The diameter in pixels of the rounded arcs.
        """
        super().__init__(width, height, x=x, y=y)

        self.corner = corner
        _platform.Platform().groundrect_constructor(self, width, height, corner)

    def __str__(self):
        # TODO(sredmond): It's a little awkward that the constructor argument order is different.
        return "GRoundRect(x={self.x}, y={self.y}, width={self.width}, height={self.height}, corner={self.corner})".format(self=self)


class G3DRect(GRect):
    """Graphical representation of a rectangular box that can be raised or lowered.

    The 3D effect of this rectangle is purely aesthetic and does not affect the
    rectangle's z-positioning on the parent :class:`GCompound`.
    """
    # TODO(sredmond): Add more documentation from the GRect object here
    def __init__(self, width, height, *, x=None, y=None, raised=False):
        """Create a new 3D rectangle with the supplied width and height.

        The caller can also specify x and y coordinates for the top-left corner
        of this :class:`G3DRect`'s bounding box.

        If this rectangle is raised (defaulting to False), then the rectangle is
        drawn with highlights that suggest that it is raised about the background.

        :param width: The width of the 3D rectangle in pixels.
        :param height: The height of the 3D rectangle in pixels.
        :param x: The x-coordinate of the top-left corner of this 3D rectangle.
        :param y: The y-coordinate of the top-left corner of this 3D rectangle.
        :param raised: Whether to draw the 3D rectangle as raised (or lowered)
        """
        super().__init__(width, height, x=x, y=y)
        self._raised = raised
        _platform.Platform().g3drect_constructor(self, width, height, raised)

    @property
    def raised(self):
        """Get or set whether this object appears raised.

        True represents a rectangle with highlights to appear raised above the background.
        False represents a rectangle with highlights to appear lowered above the background.
        """
        return self._raised

    @raised.setter
    def raised(self, is_raised):
        self._raised = is_raised
        _platform.Platform().g3drect_set_raised(self, is_raised)

    def __str__(self):
        # TODO(sredmond): It's a little awkward that the constructor argument order is different.
        return "G3DRect(x={self.x}, y={self.y}, width={self.width}, height={self.height}, raised={self.raised})".format(self=self)


class GOval(GFillableObject):
    """Graphical representation of an oval inscribed in a rectangular box.

    To display a filled green oval::

        window = GWindow()
        oval = gobjects.GOval(window.width, window.height)
        oval.filled = True
        oval.color = "GREEN"
        window.add(oval)
    """

    def __init__(self, width, height, *, x=0, y=0):
        """Initialize a new oval inscribed in a rectangular box.

        By default, the corner of the rectangular box is at (0, 0). By
        specifying x- and y- coordinates as keyword arguments, the caller can
        locate the oval at a given position.


        :param width: The width of the bounding rectangle in pixels.
        :param height: The height of the bounding rectangle in pixels.
        :param x: The x-coordinate of the top-left corner of the bounding rectangle.
        :param y: The y-coordinate of the top-left corner of the bounding rectangle.
        """
        super().__init__()
        self._width = width
        self._height = height
        # _platform.Platform().goval_constructor(self, width, height)
        # TODO(sredmond): At the moment, this needs to be after the constructor line. This is a sign of a bad organization and should be changed.
        self.location = x, y

    # def setSize(self, width=None, height=None, size=None):
    #     '''
    #     Changes the size of the bounding rectangle to the specified width
    #     and height.

    #     @type width: float
    #     @type height: float
    #     @type size: GDimension
    #     @param size: will override height and width parameters
    #     @rtype: void
    #     '''
    #     if(self._transformed): raise Exception("setSize: Object has been transformed")
    #     if(size != None):
    #         width = size.width
    #         height = size.height

    #     self.width = width
    #     self.height = height
    #     _platform.Platform().gobject_set_size(self, width, height)

    # def setBounds(self, x=None, y=None, width=None, height=None, bounds=None):
    #     '''
    #     Changes the bounds of the oval to the specified values.

    #     @type x: float
    #     @type y: float
    #     @type width: float
    #     @type height: float
    #     @type bounds: GRectangle
    #     @param bounds: will override all other parameters
    #     @rtype: void
    #     '''
    #     if(bounds != None):
    #         x, y, width, height = bounds

    #     self.setLocation(x=x, y=y)
    #     self.setSize(width, height)

    @property
    def bounds(self):
        """Get the bounding box for this :class:`GOval`.

        The bounding box of a :class:`GOval` is simply the bounding rectangle.

        :returns: A bounding box that covers this :class:`GOval`.
        :rtype: :class:`GRectangle`
        """
        # TODO(sredmond): What do we we do about transformed rectangles?
        if(self._transformed): return _platform.Platform().gobject_get_bounds(self)
        return _gtypes.GRectangle(self.x, self.y, self._width, self._height)


    def __contains__(self, point):
        """Implement ``point in self``.

        Check whether a given :class:`GPoint` or 2-element tuple is contained
        within this oval. A point is contained within the oval if it lies
        completely inside the oval or is on the boundary.

        :param point: The :class:`GPoint` or 2-element tuple to check.
        :returns: Whether the point is within this oval.
        """
        # Attempt to unpack the supplied point as a tuple.
        x, y = point

        # TODO(sredmond): How should I handle transformed GObjects?
        if self._transformed:
            return _platform.Platform().gobject_contains(self, x, y)

        rx = self.width / 2
        ry = self.height / 2
        if rx == 0 or ry == 0:  # No points in a line.
            return False

        # `point` is at (dx, dy) relative to ellipse center.
        dx = x - (self.x + rx)
        dy = y - (self.y + ry)
        return (dx * dx) / (rx * rx) + (dy * dy) / (ry * ry) <= 1.0

    def __str__(self):
        return "GOval({self.width}, {self.height}, x={self.x}, y={self.y})".format(self=self)


class GArc(GFillableObject):
    """Graphical representation of an elliptical arc.

    An elliptical arc is uniquely determined by the following three parameters:

    - The coordinates of the ellipse's bounding rectangle (width, height) and (x, y)
    - The angle at which the arc starts (start)
    - The number of degrees that the arc covers (sweep)

    All angles in the :class:`GArc` class are measured in degrees counterclockwise
    from the positive x-axis. Negative angles refer to degrees clockwise.

    To create an semicircle arc starting 45 degrees counterclockwise of the
    positive x-axis with diameter 50::

        window = GWindow()
        arc = GArc(50, 50, 45, 180, x=0, y=0)
        window.add(arc)

    If a :class:`GArc` is unfilled, the figure consists only of the arc itself.

    If a :class:`GArc` is filled, the figure consists of the
    pie-shaped wedge formed by connecting the endpoints of the arc to
    the center.

    For example, the following program draws a 270-degree arc starting at 45
    degrees filled in yellow, similar to the PacMan video game::

        window = GWindow()
        center = GPoint(window.width / 2, window.height / 2)
        radius = 25
        pacman = GArc(2 * radius, 2 * radius, 45, 270
            x=center.x - radius, y=center.y - radius)
        pacman.filled = True
        pacman.fill_color = "YELLOW"
        window.add(pacman)
    """
    def __init__(self, width, height, start, sweep, x=0, y=0):
        """Initialize a new :class:`GArc` consisting of an elliptical arc.


        By default, the corner of the rectangular box is at (0, 0). By
        specifying x- and y- coordinates as keyword arguments, the caller can
        locate the arc's bounding rectangle at a given position.

        :param width: The width of the bounding rectangle in pixels.
        :param height: The height of the bounding rectangle in pixels.
        :param start: The angle at which the arc starts (in degrees) from the positive x-axis.
        :param sweep: The angle that the arc sweeps out (in degrees).
        :param x: The x-coordinate of the top-left corner of the bounding rectangle.
        :param y: The y-coordinate of the top-left corner of the bounding rectangle.
        """
        super().__init__()
        self.location = x, y

        # What are these?
        self.frameWidth = width
        self.frameHeight = height

        self._start = start
        self._sweep = sweep
        # _platform.Platform().garc_constructor(self, width, height, start, sweep)

    @property
    def start(self):
        """Get or set the starting angle (in degrees) of this :class:`GArc`.

        The starting angle is measured in degrees counterclockwise from the
        positive x-axis.
        """
        return self._start

    @start.setter
    def start(self, start_angle):
        self._start = start_angle
        _platform.Platform().garc_set_start_angle(self, start)

    @property
    def sweep(self):
        """Get or set the sweep angle (in degrees) of this :class:`GArc`.

        The sweep angle is measured in degrees and represents the number of
        degrees carved out by this :class:`GArc`. For example, a semi-elliptical
        arc has a sweep of 180 degrees.
        """
        # TODO(sredmond): Enforce that sweep is between 0 and 360. or perhaps -360 and 360
        return self._sweep

    @sweep.setter
    def sweep(self, sweep_angle):
        self._sweep = sweep_angle
        _platform.Platform().garc_set_sweep_angle(self, sweep_angle)

    @property
    def start_point(self):
        """Get the :class:`GPoint` at which the arc starts."""
        rx, ry = self.width / 2, self.height / 2
        cx, cy = self.x + rx, self.y + ry
        return _gtypes.GPoint(cx + rx * _gmath.cos_degrees(self.start), cy - ry * math.sin_degrees(self.start))

    @property
    def end_point(self):
        """Get the :class:`GPoint` at which the arc ends."""
        rx, ry = self.width / 2, self.height / 2
        cx, cy = self.x + rx, self.y + ry
        return _gtypes.GPoint(cx + rx * _gmath.cos_degrees(self.start + self.sweep), cy - ry * math.sin_degrees(self.start + self.sweep))

    # TODO(sredmond): Expose a setter for the start point or end point?

    # def setFrameRectangle(self, x=None, y=None, width=None, height=None, rect=None):
    #     '''
    #     Changes the boundaries of the rectangle used to frame the arc.

    #     @type x: float
    #     @type y: float
    #     @type width: float
    #     @type height: float
    #     @type rect: GRectangle
    #     @param rect: bounding frame, will override other parameters
    #     @rtype: void
    #     '''
    #     if(rect != None):
    #         x = rect.getX()
    #         y = rect.getY()
    #         width = rect.getWidth()
    #         height = rect.getHeight()

    #     self.location = x, y
    #     self.frameWidth = width
    #     self.frameHeight = height
    #     _platform.Platform().garc_set_frame_rectangle(self, x, y, width, height)

    # def getFrameRectangle(self):
    #     '''
    #     Returns the boundaries of the rectangle used to frame the arc.

    #     @rtype: GRectangle
    #     '''
    #     return _gtypes.GRectangle(0,0,0,0) #!!

    @property
    def bounds(self):
        """Get the bounding rectangle for this arc.

        @rtype: GRectangle
        """
        # TODO(sredmond): Should this be small if the arc is small?
        # TODO(sredmond): Write this method.
        if(self._transformed): return _platform.Platform().gobject_get_bounds(self)

        rx = self.frameWidth / 2
        ry = self.frameHeight / 2
        cx = self.x + rx
        cy = self.y + ry

        startRadians = self.start * _gmath.PI / 180
        sweepRadians = self.sweep * _gmath.PI / 180

        p1x = cx + math.cos(startRadians) * rx
        p1y = cy - math.sin(startRadians) * ry
        p2x = cx + math.cos(startRadians + sweepRadians) * rx
        p2y = cy - math.sin(startRadians + sweepRadians) * ry

        xMin = min(p1x, p2x)
        xMax = max(p1x, p2x)
        yMin = min(p1y, p2y)
        yMax = max(p1y, p2y)

        if self.containsAngle(0): xMax = cx + rx
        if self.containsAngle(90): yMin = cy - ry
        if self.containsAngle(180): xMin = cx - rx
        if self.containsAngle(270): yMax = cy + ry

        if self.filled:
            xMin = min(xMin, cx)
            yMin = min(yMin, cy)
            xMax = max(xMax, cx)
            yMax = max(yMax, cy)
        return _gtypes.GRectangle(xMin, yMin, xMax - xMin, yMax - yMin)

    def __contains__(self, point, tolerance=2.5):
        """Implement ``point in self``.

        A :class:`GArc` contains a point if and only if the point is contained
        inside of the pie wedge defined by the bounds of the arc.

        # For advanced users, this method is overloaded with an optional tolerance
        # parameter. Since *almost* no one will ever want to directly call
        # ``self.__contains__(point)``, the tolerance defaults to a reasonable value.
        # However, to check for containment with a custom tolerance, call::

        #     line = GLine(0, 0, 5, 5)
        #     print(line.__contains__((2, 2), tolerance=0.5))  # => False

        # :param point: The :class:`GPoint` or 2-element tuple to check.
        # :param tolerance: The maximum distance (in pixels) to still count as containment.
        # :returns: Whether the point is within a tolerance of any point on this line.
        """

        '''
        Returns whether or not this object contains the point x, y

        @type x: float
        @type y: float
        @rtype: boolean
        '''
        # TODO(sredmond): Write this method correctly.
        if(self._transformed): return _platform.Platform().gobject_contains(self, x, y)

        rx = frameWidth / 2
        ry = frameHeight / 2
        if rx == 0 or ry == 0: return False
        dx = x - (self.x + rx)
        dy = y - (self.y + ry)
        r = (dx * dx) / (rx * rx) + (dy * dy) / (ry * ry)
        if(self.fillFlag):
            if (r > 1.0): return False
        else:
            t = tolerance / ((rx + ry) / 2) #!!
            if (abs(1.0 - r) > t): return False
        return self.containsAngle(math.atan2(-dy, dx) * 180 / _gmath.PI)

    def containsAngle(self, theta):
        '''
        Internal method
        '''
        start = min(self.start, self.start + self.sweep)
        sweep = abs(self.sweep)
        if(sweep >= 360): return True
        if(theta < 0): theta = 360 - math.fmod(-theta, 360)
        else: theta = math.fmod(theta, 360)
        if(start < 0): start = 360 - math.fmod(-start, 360)
        else: start = math.fmod(start, 360)
        if(start + sweep > 360):
            return (theta >= start or theta <= start + sweep - 360)
        else:
            return (theta >= start and theta <= start + sweep)

    def __str__(self):
        return "GArc({self.frame_width}, {self.frame_height}, {self.start}, {self.sweep}, x={self.x}, y={self.y}".format(self=self)


class GLine(GObject):
    """Graphical representation of a line segment defined by two endpoints.

    To create a line from ``(14, 41)`` to ``(601, 106)``::

        line = GLine(14, 41, 601, 106)

    To draw two diagonal lines crossing a :class:`GWindow`::

        window = GWindow()
        window.add(GLine(0, 0, window.width, window.height))
        window.add(GLine(0, window.height, window.width, 0))

    Note that creating a :class:`GLine` will not automatically add it to any
    displays. Rather, you must explicitly call the `.add` method on an
    appropriate :class:`GWindow`.
    """
    def __init__(self, x0, y0, x1, y1):
        """Create a line segment from its endpoints.

        The first two parameters ``(x0, y0)`` define the start of the line, and
        the next two parameters ``(x1, y1)`` define the end of the line.

        To create a line from ``(0, 0)`` to ``(41, 106)``::

            line = GLine(0, 0, 41, 106)

        :param x0: The x-coordinate of the starting point.
        :param y0: The y-coordinate of the starting point.
        :param x1: The x-coordinate of the ending point.
        :param y1: The y-coordinate of the ending point.
        """
        # TODO(sredmond): Consider allowing the user to supply two GPoints (with splats?)
        super().__init__()
        self._x0 = x0
        self._y0 = y0
        self._x1 = x1
        self._y1 = y1

        # TODO(sredmond): Don't bother sending along the values when we also send the built object.
        # _platform.Platform().gline_constructor(self)

    @property
    def start(self):
        """Get or set this :class:`GLine`'s starting point.

        Changing the start point modifies the line segment even if it has
        already been added to a :class:`GWindow`.

        Usage::

            window = GWindow()
            line = GLine(0, 0, 41, 41)
            window.add(line)
            print(line.start)  # => GPoint(0, 0)
            line.start = 4, 1
            print(line.start)  # => GPoint(4, 1)

        You can supply either a GPoint or a 2-element tuple to this property.
        Accessing this property will return a GPoint which can be unpacked as a
        tuple::

            window = GWindow()
            line = GLine(0, 0, 41, 41)
            window.add(line)
            line.start = GPoint(16, 25)  # Supply a GPoint to the setter.
            startx, starty = line.start  # Unpack the start point immediately.

        Setting the start point does not modify the end point. Thus, this is
        different from setting this :class:`GLine`'s location, which translates
        the entire line segment.
        """
        return _gtypes.GPoint(self._x0, self._y0)

    @start.setter
    def start(self, start_point):
        # Attempt to unpack the supplied start point as a tuple. This supports
        # both GPoints and 2-element tuples.
        start_x, start_y = start_point
        self._x0 = start_x
        self._y0 = start_y
        _platform.Platform().gline_set_start_point(self, self._x0, self._y0)

    @property
    def end(self):
        """Get or set this :class:`GLine`'s ending point.

        Changing the end point modifies the line segment even if it has
        already been added to a :class:`GWindow`.

        Usage::

            window = GWindow()
            line = GLine(0, 0, 41, 41)
            window.add(line)
            print(line.end)  # => GPoint(41, 41)
            line.end = 4, 1
            print(line.end)  # => GPoint(4, 1)

        You can supply either a GPoint or a 2-element tuple to this property.
        Accessing this property will return a GPoint which can be unpacked as a
        tuple::

            window = GWindow()
            line = GLine(0, 0, 41, 41)
            window.add(line)
            line.end = GPoint(106, 106)  # Supply a GPoint to the setter.
            endx, endy = line.start  # Unpack the end point immediately.

        Setting the end point does not modify the start point. Thus, this is
        different from setting this :class:`GLine`'s location, which translates
        the entire line segment.
        """
        return _gtypes.GPoint(self._x1, self._y1)

    @end.setter
    def end(self, end_point):
        # Attempt to unpack the supplied start point as a tuple. This supports
        # both GPoints and 2-element tuples.
        end_x, end_y = start_point
        self._x1 = end_x
        self._y1 = end_y
        _platform.Platform().gline_set_start_point(self, self._x1, self._y1)

    # TODO(sredmond): Add methods to get/set dx/dy?
    # TODO(sredmond): Add methods to get/set just x0/y0/x1/y1?

    # TODO(sredmond): Implement bounds.
    # @property
    # def bounds

    def __contains__(self, point, tolerance=1.5):
        """Implement ``point in self``.

        Check whether a given :class:`GPoint` or 2-element tuple is contained
        within this line.

        Since a line segment technically has zero area, this method actual checks
        whether the supplied point is within some small, default tolerance (inclusive)
        of any point on the line segment, measured by perpendicular distance.
        For example, the point ``(2, 2)`` is within the default tolerance of
        the line from ``(0, 0)`` to ``(5, 5)``::

            line = GLine(0, 0, 5, 5)
            print((2, 2) in line)  # => True
            print((0, 5) in line)  # => False

        For advanced users, this method is overloaded with an optional tolerance
        parameter. Since *almost* no one will ever want to directly call
        ``self.__contains__(point)``, the tolerance defaults to a reasonable value.
        However, to check for containment with a custom tolerance, call::

            line = GLine(0, 0, 5, 5)
            print(line.__contains__((2, 2), tolerance=0.5))  # => False

        :param point: The :class:`GPoint` or 2-element tuple to check.
        :param tolerance: The maximum distance (in pixels) to still count as containment.
        :returns: Whether the point is within a tolerance of any point on this line.
        """
        # TODO(sredmond): This method will need to change if the GLine is transformed.

        # The distance between a point (x, y) and a line (infinite, not a segment)
        # between two points (x0, y0) and (x1, y1) is given by:
        #
        #    |(y1 - y0) * x - (x1 - x0) * y + x1 * y0 - y1 * x0|
        #    ---------------------------------------------------
        #           distance between (x0, y0) and (x1, y1)
        # Attempt to unpack the supplied point as a tuple.
        x, y = point

        # If our line segment is really a point, just check the circle.
        if self._x0 == self._x1 and self._y0 == self._y1:
            return math.hypot(self._x0 - x, self._y0 - y) <= tolerance

        # If the distance to either of the endpoints is small enough, we're good.
        if math.hypot(self._x0 - x, self._y0 - y) <= tolerance or math.hypot(self._x1 - x, self._y1 - y) <= tolerance:
            return True

        # Compute coefficients a, b, c for the expression of this line as ax + by + c = 0
        a = self._y1 - self._y0
        b = self._x0 - self._x1
        c = self._x1 * self._y0 - self._y1 * self._x0
        dist = math.hypot(a, b)

        proj_x = (b * (b * x - a * y) - a * c) / (dist * dist)
        proj_y = (a * (-b * x + a * y) - b * c) / (dist * dist)

        # Otherwise, we're only close enough if (1) the point we project to (on
        # the full line) is between the two endpoints AND (2) we are close enough
        # to that point.
        return (min(self._x0, self._x1) <= proj_x <= max(self._x0, self._x1)
            and min(self._y0, self._y1) <= proj_y <= max(self._y0, self._y1)
            and abs(a * x + b * y + c) / dist <= tolerance)

    def __str__(self):
        return "GLine({self.x0}, {self.y0}, {self.x1}, {self.y1}".format(self=self)

class GImage(GObject):
    """Graphical representation of an image from a file.

    To display a centered :class:`GImage` of the Stanford tree::

        window = GWindow()
        tree = GImage("StanfordTree.gif")
        x = (window.width - tree.width) / 2
        y = (window.height - tree.height) / 2
        window.add(tree, x, y)

    The above code assumes that you have a file named ``StanfordTree.gif`` in
    the current directory.

    The search for matching image files begins in the current directory and,
    failing that, searches an ``images/`` subdirectory relative to the calling script.
    """
    # TODO(sredmond): Have a environmental variable for a custom subdirectory.
    # TODO(sredmond): When would you use a GImage over a GBufferedImage?

    def __init__(self, filename, x=0, y=0):
        """Initialize a new image from the given file.

        The image file should either exist in the current directory or in a
        subdirectory named ``images/``.

        By default, the upper left corner of the image appears at the origin.
        By supplying x- and y- coordinates, the caller can set the location
        to the point (x, y).

        :param filename: the name of the image file to load.
        :param x: (optional) the x-coordinate to which to move this :class:`GImage`.
        :param y: (optional) the y-coordinate to which to move this :class:`GImage`.
        """
        super().__init__()
        self._path = _platform.Platform().image_find(filename)
        self._image = _platform.Platform().gimage_constructor(self)
        self.location = x, y
        self._size = _platform.Platform().gimage_constructor(self, filename)

    @property
    def filename(self):
        """Get this :class:`GImage`'s filename."""
        return self._path.name

    @property
    def bounds(self):
        """Get the bounding box for this :class:`GImage`."""
        if self._transformed: return _platform.Platform().gobject_get_bounds(self)
        return _gtypes.GRectangle(self.x, self.y, self._size.width, self._size.height)

    def __str__(self):
        return 'GImage({!r})'.format(self.filename)


class GLabel(GObject):
    """Graphical representation of a text string.

    For example, to add a :class:`GLabel` with the message "Hello, world!" to
    the center of a :class:`GWindow`::

        window = GWindow()
        label = GLabel("Hello, world!")
        label.font = "SansSerif-18"
        x = (window.width - label.width) / 2
        y = (window.height + label.ascent) / 2
        window.add(label, x, y)

    In order to control the appearance and positioning of a :class:`GLabel`,
    it's important to understand the following terms:

        - The baseline is the horizontal line on which the characters rest.
        - The origin is the point on the baseline at which the label begins.
        - The height is the distance (in pixels) that would separate two
          successive lines such that characters would not overlap.
        - The ascent is the maximum distance that any character in this font
          extends above the baseline.
        - The descent is the maximum distance that any character in this font
          extends below the baseline.
    """
    # TODO(sredmond): Are ascent and descent distances for the given text, or for the font itself?
    # TODO(sredmond): What exactly do we mean in the description of height above?

    def __init__(self, label, x=0, y=0):
        """Initialize a :class:`GLabel` displaying a given label.

        Supplying x and y arguments sets the location of the :class:`GLabel` to
        (x, y). Recall that the location of a :class:`GLabel` is its bottom-left
        coordinate.

        :param label: The message to display in this :class:`GLabel`.
        :param x: The x-coordinate of the bottom-left corner of this rectangle.
        :param y: The y-coordinate of the bottom-left corner of this rectangle.
        """
        super().__init__()
        self._text = label
        self.font = _gfont.GFont.default()
        self.location = x, y

        # TODO(sredmond): Check that overriding bounds makes width and height work appropriately.
        # size = _platform.Platform().glabel_get_size(self)
        # self.width = size.width
        # self.height = size.height


    @property
    def font(self):
        """Get or set this :class:`GLabel`'s font."""
        return self._font

    @font.setter
    def font(self, font):
        if not isinstance(font, _gfont.GFont):
            font = _gfont.GFont.parse(font)

        self._font = font
        _platform.Platform().glabel_set_font(self, font)

        # size = _platform.Platform().glabel_get_size(self)
        # self.width = size.width
        # self.height = size.height

    @property
    def text(self):
        """Get or set this :class:`GLabel`'s label.

        Setting the label changes the internal state of the :class:`GLabel`,
        so a existing drawn labels will redisplay."""
        return self._text

    @text.setter
    def text(self, text):
        self._text = text
        _platform.Platform().glabel_set_label(self, text)

        # size = _platform.Platform().glabel_get_size(self)
        # self.width = size.width
        # self.height = size.height

    @property
    def ascent(self):
        """Return the maximum distance strings in this font ascend above the baseline."""
        return self.font.ascent

    @property
    def descent(self):
        """Return the maximum distance strings in this font descend below the baseline."""
        return self.font.descent

    @property
    def bounds(self):
        """Get the bounding box for this :class:`GLabel`.

        The bounding box of a :class:`GLabel` is the smallest rectangle that
        completely covers the displayed message.

        :returns: A bounding box that covers this :class:`GLabel`'s message.
        :rtype: :class:`GRectangle`
        """
        # TODO(sredmond): How do we handle bounds when there could be descent?

        if self._transformed: return _platform.Platform().gobject_get_bounds(self)
        width = self.font.measure(self.text)
        height = self.font.linespace
        # TODO(sredmond): This assumes that linespace is ascent + descent, and we top-justify
        # the bounding box so the actual box's y-coordinates go from y - ascent to y - ascent + height.
        return _gtypes.GRectangle(self.x, self.y - self.ascent, width, height)

    def __str__(self):
        return 'GLabel("{}")'.format(self.label)


class GPolygon(GFillableObject):
    """Graphical representation of a polygon.

    Polygons are bounded by line segments. A :class:`GPolygon` is initialized
    to an empty polygon with no vertices. To complete the figure, you need to
    add vertices to the polygon using the methods :meth:`add_vertex`,
    :meth:`add_edge`, or :meth:`add_polar_edge`.

    To add a filled red octagon to the center of the window::

        window = GWindow()
        edge_length = 75
        stop_sign = GPolygon()
        start = GPoint(-edge_length / 2, edge_length / 2 + edge_length / math.sqrt(2.0))
        stop_sign.add_vertex(start)
        for edge in range(8):
            stop_sign.add_polar_edge(edge_length, 45*edge)
        stop_sign.filled = True
        stop_sign.color = "RED"
        window.add(stop_sign, window.width / 2, window.height / 2)
    """

    def __init__(self):
        """Initialize an empty polygon at the origin."""
        super().__init__()
        self.last_x = 0
        self.last_y = 0
        self.vertices = []
        # _platform.Platform().gpolygon_constructor(self)

    def add_vertex(self, point):
        """Add a vertex at a given :class:`GPoint` relative to the polygon origin."""
        # TODO(sredmond): Flush out this documentation with the usual messages.
        x, y = point
        self.last_x = x
        self.last_y = y
        self.vertices.append(_gtypes.GPoint(x, y))
        _platform.Platform().gpolygon_add_vertex(self, x, y)

    def add_edge(self, dx, dy):
        """
        Adds an edge to the polygon whose components are given by the displacements
        dx and dy from the last vertex.
        """
        # TODO(sredmond): Fix documentation
        self.add_vertex((self.last_x + dx, self.last_y + dy))

    def add_polar_edge(self, r, theta):
        """
        Adds an edge to the polygon specified in polar coordinates.  The length
        of the edge is given by r, and the edge extends in
        direction theta, measured in degrees counterclockwise
        from the +x axis.

        @type r: float
        @type theta: float
        @param theta: degrees
        @rtype: void
        """
        # TODO(sredmond): Fix documentation
        self.add_edge(
            r * _gmath.cos_degrees(theta),
            -r * _gmath.sin_degrees(theta)
        )

    @property
    def bounds(self):
        """Get the bounding box for this :class:`GPolygon`.

        :returns: The bounding box of this :class:`GPolygon`.
        :rtype: :class:`GRectangle`
        """
        if self._transformed: return _platform.Platform().gobject_get_bounds(self)

        xs = [vertex.x for vertex in self.vertices]
        ys = [vertex.y for vertex in self.vertices]
        min_x = min(xs)
        max_x = max(xs)
        min_y = min(ys)
        max_y = max(ys)
        return _gtypes.GRectangle(min_x, min_y, max_x - min_x, max_y - min_y)


    def __contains__(self, point):
        """Implement ``point in self``.
        """
        # TODO(sredmond): Add the usual documentation here.
        x, y = point
        if self._transformed: return _platform.Platform().gobject_contains(self, x, y)

        # TOOD(sredmond): Verify this containment algorithm.
        crossings = 0

        n = len(self.vertices)
        if(n < 2): return False
        if(vertices[0] == vertices[n-1]): n = n - 1
        x0 = vertices[0].getX()
        y0 = vertices[0].getY()
        for i in range(1, n+1):
            x1 = vertices[i % n].getX()
            y1 = vertices[i % n].getY()
            if((y0 > y) != (y1 > y) and x - x0 < (x1-x0)*(y-y0)/(y1-y0)):
                crossings = crossings + 1
            x0 = x1
            y0 = y1
        return (crossings % 2 == 1)

    def __iter__(self):
        return iter(self.vertices)

    def __str__(self):
        return "GPolygon(num_vertices={})".format(len(self.vertices))


class GCompound(GObject, MutableSequence):
    """A graphics object that is a collection of other graphics objects.

    Once assembled, the contained :class:`GObject`s can be manipulated as a unit.

    The :class:`GCompound` has its own position, and items within in are drawn
    relative to that position.

    Internally, the :class:`GCompound` just holds a stack of its :class:`GObject`s.
    """
    # TODO(sredmond): I was a little delirious when I wrote this - take another look over it.
    # TODO(sredmond): It's a little weird to make this into a collection.
    def __init__(self):
        """Create an empty :class:`GCompound`."""
        super().__init__()
        self.contents = []
        _platform.Platform().gcompound_constructor(self)

    # These abstract methods end up adding a bunch of derived methods, some of which are nice.
    # One of these is __contains__, which we override for the point containment. That might break
    # something unexpectedly, so TODO(sredmond): Read more about these abstract superclasses and
    # make a call.
    def __getitem__(self, index):
        """Return the :class:`GObject` at the given index.

        The number is from back to front in the z-dimension. That is, index 0
        refers to the backmost element and index -1 refers to the frontmost element.
        """
        return self.contents[index]

    def __setitem__(self, index, value):
        self.contents[index] = value

    def __delitem__(self, index):
        gobj = self.contents.pop(index)
        _platform.Platform().gobject_remove(gobj)
        gobj._parent = None

    def __len__(self):  # Definitely keep this one!
        """Return the number of graphical objects stored in this :class:`GCompound`."""
        return len(self.contents)

    def insert(self, index, value):
        self.contents.insert(index, value)
    # End abstract methods.

    def add(self, gobj, x=None, y=None):
        """Add a new :class:`GObject` to this :class:`GCompound`.

        If two additional arguments x and y are both supplied, move the object
        to ``(x, y)`` first. It is an error to specify just one of x and y.

        :param gobj: The object to add to this compound.
        :param x: (optional) The x-coordinate of the location to which to move this object.
        :param y: (optional) The y-coordinate of the location to which to move this object.
        """
        # TODO(sredmond): Raise an error if only x or only y is set.
        if x is not None and y is not None:
            gobj.location = (x, y)

        # Dispatch the constructor to the appropriate backend constructor.
        if isinstance(gobj, GLine):
            _platform.Platform().gline_constructor(gobj)

        if isinstance(gobj, GImage):
            # TODO(sredmond): Warn against creating a polygon w/o enough vertices.
            _platform.Platform().gimage_constructor(gobj)

        if isinstance(gobj, GRect):
            _platform.Platform().grect_constructor(gobj)

        if isinstance(gobj, GOval):
            _platform.Platform().goval_constructor(gobj)

        if isinstance(gobj, GArc):
            _platform.Platform().garc_constructor(gobj)

        if isinstance(gobj, GLabel):
            _platform.Platform().glabel_constructor(gobj)

        if isinstance(gobj, GPolygon):
            # TODO(sredmond): Warn against creating a polygon w/o enough vertices.
            _platform.Platform().gpolygon_constructor(gobj)

        _platform.Platform().gcompound_add(self, gobj)
        self.contents.append(gobj)
        gobj._parent = self

    def remove(self, gobj):
        """Remove a :class:`GObject` from this :class:`GCompound`.

        Return whether or not the :class:`GCompound` contained the :class:`GObject`.

        :param gobj: The object to remove.
        :returns: Whether this compound contained the object.
        """
        # TODO(sredmond): People familiar with Python lists might expect this to raise a ValueError if the gobj isn't there.

        # TODO(sredmond): Don't blindly delete from the backend if it's not added.
        _platform.Platform().gobject_remove(gobj)

        try:
            self.contents.remove(gobj)
        except ValueError:
            return False
        else:
            return True

    def clear(self):
        """Remove all graphical objects from the GCompound."""
        while self.contents:
            self.pop()

    @property
    def bounds(self):
        """Get the bounding box for this :class:`GCompound`.

        The bounding box of a :class:`GCompound` is the smallest rectangle that
        covers all of its components bounding boxes.

        :returns: A bounding box that covers all of this compounds components.
        :rtype: :class:`GRectangle`
        """
        if self._transformed:
            return _platform.Platform().gobject_get_bounds(self)

        min_x = min((obj.x for obj in self.contents), default=-1)
        min_y = min((obj.y for obj in self.contents), default=-1)
        max_x = max((obj.x + obj.width for obj in self.contents), default=-1)
        max_y = max((obj.y + obj.height for obj in self.contents), default=-1)

        return _gtypes.GRectangle(min_x, min_y, max_x - min_x, max_y - min_y)

    def __contains__(self, point):
        """Implement ``point in self``.

        Check whether a given :class:`GPoint` or 2-element tuple is contained
        within the bounding box of this :class:`GCompound`. A point is contained
        within a GCompound if it is contained within any of the components.
        """
        # Attempt to unpack the supplied point as a tuple.
        x, y = point

        if self._transformed:
            return _platform.Platform().gobject_contains(self, x, y)

        return any((x, y) in obj for obj in self.contents)

    def send_forward(self, gobj):
        try:
            index = self.contents.find(gobj)
        except ValueError:
            return
        else:
            if index != len(self) - 1:
                self.contents.pop(index)
                self.contents.insert(index + 1, gobj)
                _platform.Platform().gobject_send_forward(gobj)

    def send_to_front(self, gobj):
        try:
            index = self.contents.find(gobj)
        except ValueError:
            return
        else:
            if index != len(self) - 1:
                self.contents.pop(index)
                self.contents.append(index + 1, gobj)
                _platform.Platform().gobject_send_to_front(gobj)

    def send_backward(self, gobj):
        try:
            index = self.contents.find(gobj)
        except ValueError:
            return
        else:
            if index > 0:
                self.contents.pop(index)
                self.contents.insert(index - 1, gobj)
                _platform.Platform().gobject_send_backward(gobj)

    def send_to_back(self, gobj):
        try:
            index = self.contents.find(gobj)
        except ValueError:
            return
        else:
            if index > 0:
                self.contents.pop(index)
                self.contents.insert(0, gobj)
                _platform.Platform().gobject_send_to_back(gobj)

    def __iter__(self):
        return iter(self.contents)



"""Manage initialization of and communications to and from the Java backend.

The Java backend should live in a JAR file adjacent to this module.

The backend is initialized lazily, only created once someone tries to write
to or read from the pipe. This means that the first call to the pipe will be
slow, as it must take time to spawn the Java process.
"""
# TODO(sredmond): Consider updating the subprocess calls to use Python 3.5+
import logging
import pathlib
import queue  # TODO(sredmond): Double-check that this doesn't have to be multiprocessing.queue
import shlex
import subprocess
import sys
import threading

from campy.system import error

# Module-level logger.
logger = logging.getLogger(__name__)


# The backend JAR should live adjacent to this file.
SPL_JAR_LOCATION = pathlib.Path(__file__).parent / 'spl.jar'


# Use `shlex` to parse shell text. These args, when passed to `Popen`, launch the Java backend.
LAUNCH_SPL_ARGS = shlex.split('java -jar {}'.format(SPL_JAR_LOCATION))


def debug_print(line):
    # if DEBUG_PIPE:
    # options['file'] = sys.stderr
    # options['flush'] = True

    logger.info(line)
    # print(*values, **options)


def append_output_to_queue(out, queue):
    for line in iter(out.readline, b''):
        logger.info('Adding output to queue: %r', line)
        queue.put(line)
    out.close()


class JavaBackendPipe:
    """A JavaBackendPipe connects to a single instance of a running JAR file.

    A JavaBackendPipe knows how to write strings to the backend and how to read
    responses from the backend.

    In order to avoid side-effects on import, the JAR is not run until a read or
    write is requested.

    Implementation note: A JavaBackendPipe starts a thread to consume the output
    from the Java backend, so that the internal OS pipe is unlikely to fill up.
    When the JavaBackendPipe is deleted, the thread is joined as well.
    """

    def __init__(self):
        self._pipe = None

    @property
    def pipe(self):
        # TODO(sredmond): Adding a level of indirection here (self.pipe calls a
        # function) seems like it could slow down the frequent operation of
        # using this pipe. Perhaps consider a different way to make pipe lazily-
        # initialized.
        if not self._pipe:
            # TODO(sredmond): Double check that these are the correct kwargs.
            # TODO(sredmond): Crash if the SPL can't be found or if this process exits
            # early.
            self._pipe = subprocess.Popen(LAUNCH_SPL_ARGS,
                                           shell=False,
                                           stdin=subprocess.PIPE, \
                                           stdout=subprocess.PIPE, \
                                           stderr=sys.stdout, \
                                           universal_newlines=True)

            self.line_queue = queue.Queue()
            self.watchdog = threading.Thread(target=append_output_to_queue,
                                             args=(self._pipe.stdout, self.line_queue))
            self.watchdog.daemon = True # thread dies with the program
            self.watchdog.start()

        return self._pipe

    def write(self, line):
        line += '\n'  # TODO(sredmond): Do we only need to add a newline when
        # there isn't already one, or always?
        # TODO(sredmond): Consider self.pipe.communicate(input=line, timeout=1)[0]
        logger.info('Writing line: %r', line)
        debug_print(line)
        self.pipe.stdin.write(line)
        self.pipe.stdin.flush()

    def read(self):
        logger.info('Queue size: %d', self.line_queue.qsize())
        return self.line_queue.get()
        # Alternatively, return self.pipe.stdout.readline()

    def get_status(self):
        result = self.get_result()
        if result != 'ok':
            error(result)

    # TODO: check for whitespace returned at start or finish
    def get_result(self, consume_acks=True, stop_on_event=False, caller=''):
        while True:
            debug_print('get_result(): calling read()...')
            line = self.read()
            debug_print(line)

            is_result = line.startswith('result:')
            is_result_long = line.startswith('result_long:')
            is_event = line.startswith('event:')
            is_ack = line.startswith('result:___jbe___ack___')
            has_acm_exception = 'acm.util.ErrorException' in line
            has_exception = 'xception' in line
            has_error = 'Unexpected error' in line

            if is_result_long:
                # Read a long result (sent across multiple lines)
                result = ''
                next_line = self.read()
                while next_line != 'result_long:end':
                    if not line.startswith('result:___jbe___ack___'):
                        result += line
                        debug_print('getResult(): appended line (length so far: {})'.format(len(result)))
                    next_line = self.read()
                debug_print('getResult(): returning long strings "{}...{}" (length {})'.format(result[:10], result[-10:], len(result)))
                return result
            elif ((is_result or is_event) and has_acm_exception) or (not is_result and not is_event and (has_exception or has_error)):
                # Read an error message from the back-end
                if is_result:
                    line = line[7:]  # Prune 'result:'
                elif is_event:
                    line = line[6:]  # Prune 'event:'
                result = 'ERROR emitted from Stanford Java back-end process\n{}'.format(line)
                error(result)

            elif is_result:
                # Read a regular result
                if not is_ack or not consume_acks:
                    result = line[7:]  # Prune 'result:'
                    debug_print('getResult(): returning regular result (length {}) {}'.format(len(result), repr(result)))
                    return result.strip()
                else:
                    # Just an acknowledgement of some previous event: not a real result.
                    debug_print('getResult(): saw ACK (length {}) "{}"'.format(len(line), line))
            elif is_event:
                # Read a Java-originated event; enqueue it to process here.
                event = self.parseEvent(line[6:].strip())
                Platform.EVENT_QUEUE.append(event)
                if stop_on_event or (event.event_class == gevents.EventClassType.WINDOW_EVENT and event.event_type == gevents.EventType.CONSOLE_CLOSED and caller == 'get_line_console'):
                    return ''
            else:
                if '\tat ' in line or '   at ' in line:
                    # a line from a back-end Java exception stack trace;
                    # shouldn't really be happening, but back end isn't perfect.
                    # echo it here to STDERR so Python user can see it to help diagnose the issue
                    debug_print(line)

    def __del__(self):
        self._pipe.terminate()
        # TODO(sredmond): Do we want to join threads now or at program exit?
        # pass




"""

"""
import collections as _collections

class LinkedList(_collections.abc.MutableSequence):
    def __init__(self):
        self.head = None

    # __getitem__, __setitem__, __delitem__, __len__, insert

    def add(self):
        pass

class LinkedListNode(object):
    def __init__(self, value, next_node):
        self.value = value
        self.next = next_node

    def __str__(self):
        return "LLNode(value={self.value})".format(self=self)



"""
Exports a :class:`Note` class that can play musical notes.

A :class:`Note` has several attributes - its :class:`Pitch`

To create a note, you can
"""
import enum
import logging

import campy.private.platform as _platform

# Create a module-level logger.
logger = logging.getLogger(__name__)


# TODO: replace w/ imported error
def error(msg):
    import sys
    print(msg)
    sys.exit(1)


# Minimum legal value that an octave can have.
OCTAVE_MIN = 1


# Maximum legal value that an octave can have.
OCTAVE_MAX = 10


class Note:
    @enum.unique
    class Pitch(enum.Enum):
        R = 0  # Rest
        A = 1
        B = 2
        C = 3
        D = 4
        E = 5
        F = 6
        G = 7

    @enum.unique
    class Accidental(enum.Enum):
        SHARP = 1
        NATURAL = 0
        FLAT = -1

    def __init__(self, duration, pitch, octave, accidental=Accidental.NATURAL, repeat=False):
        self.duration = duration
        self.pitch = pitch
        self.octave = octave
        self.accidental = accidental
        self.repeat = repeat

    @classmethod
    def from_line(cls, line):
        """Constructs a new note from a line definition.

        Usage:
            note = Note.from_line('1.5 G 5 NATURAL false')
        """
        try:
            duration, pitch, octave, accidental, repeat = line.split(' ')
        except ValueError as exc:
            error(exc)

        try:
            duration = int(duration)
        except ValueError:
            error('Invalid duration numeric format.')

        try:
            pitch = Note.Pitch[pitch]
        except KeyError:
            error('Illegal pitch.')

        try:
            octave = int(octave)
        except ValueError:
            error('Illegal octave numeric format.')

        try:
            accidental = Note.Accidental[accidental]
        except KeyError:
            error('Illegal accidental.')

        repeat = repeat not in ['False', 'false']  # Compatibility with CPP-style casing

        return cls(duration, pitch, octave, accidental, repeat)

    @classmethod
    def rest(cls, duration, repeat=False):
        # TODO(sredmond): Do we ever want to allow a repeating rest? What does that even mean?
        print(duration, Note.Pitch.R, OCTAVE_MIN, Note.Accidental.NATURAL, False)
        print(cls)
        return cls(duration, Note.Pitch.R, OCTAVE_MIN, Note.Accidental.NATURAL, repeat=repeat)

    def is_rest(self):
        return self.pitch == Note.Pitch.R

    @property
    def duration(self):
        return self._duration

    @property
    def pitch(self):
        return self._pitch

    @property
    def octave(self):
        return self._octave

    @property
    def accidental(self):
        return self._accidental

    @duration.setter
    def duration(self, duration):
        if duration < 0:
            error('Illegal negative duration.')
        self._duration = duration

    @pitch.setter
    def pitch(self, pitch):
        # TODO(sredmond): Should we ever allow a rest to turn into a non-rest or vice versa?
        if pitch not in Note.Pitch:
            error('Illegal pitch.')
        self._pitch = pitch

    @octave.setter
    def octave(self, octave):
        if not OCTAVE_MIN <= octave <= OCTAVE_MAX:
            error('Illegal octave.')
        # TODO(sredmond): What the heck is this doing in here? Why would we want to increment the octave of a rest?
        # if self.is_rest():
        #     octave += 1
        self._octave = octave

    @accidental.setter
    def accidental(self, accidental):
        if accidental not in Note.Accidental:
            error('Illegal accidental value.')
        if self.is_rest():
            accidental = Note.Accidental.NATURAL
        self._accidental = accidental

    def play(self):
        logger.warning('Playing %s', self)
        _platform.Platform().note_play(self, self.repeat)

    def __str__(self):
        out = "{duration} {pitch} ".format(duration=self.duration, pitch=self.pitch.name)
        if not self.is_rest():
            out += "{octave} {accidental} ".format(octave=self.octave, accidental=self.accidental.name)
        return out


__all__ = ['OCTAVE_MIN', 'OCTAVE_MAX', 'Note']




"""Provide convenient top-level imports for the :mod:`campy.graphics` subpackage.

Library code should still use absolute imports where possible. That is, prefer::

    from campy.graphics.gcolor import GColor

more than::

    from campy.graphics import GColor

These overrides are provided for student (and debugging!) convenience. They are
subject to be removed at any time in the future and should not be considered
part of the API.
"""
# from campy.graphics.gbufferedimage import GBufferedImage
from campy.graphics.gcolor import GColor, Pixel
# GEvents?
# from campy.graphics.gfilechooser import show_open_dialog, show_save_dialog
from campy.graphics.gmath import (
    PI, E,
    to_degrees, to_radians,
    sin_degrees, cos_degrees, tan_degrees,
    vector_magnitude, vector_angle,
    count_digits
)
# from campy.graphics.gobjects import (
#     GObject,
#     GRect, GRoundRect, G3DRect,
#     GOval, GArc,
#     GLine,
#     GImage,
#     GLabel,
#     GPolygon,
#     GCompound
# )
from campy.graphics.goptionpane import ConfirmType, ConfirmResult, MessageType, GOptionPane
from campy.graphics.gtimer import GTimer
from campy.graphics.gtypes import GPoint, GDimension, GRectangle
from campy.graphics.gwindow import GWindow



import collections as _collections
import functools as _functools

@_functools.total_ordering
class SparseGrid(_collections.abc.MutableMapping):
    def __init__(self, rows=0, cols=0, value=None):
        self.rows = rows
        self.cols = cols
        self._elements = _collections.defaultdict(dict)
        if value is not None:  # Even if user supplies an empty string, we want to fill
            self.fill(value)

    @classmethod
    def from_2d_iterable(cls):
        pass

    def resize():
        pass

    def __getitem__(self, key):
        if isinstance(key, tuple):
            # Invoked like grid[row, col]
            row, col = key
            self.check_indexes(row, col);
            if row not in self._elements:
                raise KeyError(key)
            if col not in self._elements[row]:
                raise KeyError(key)
            return self._elements[row][col]
        else:
            # Invoked like grid[row][col]
            if key not in self._elements:
                raise KeyError(key)
            return self._elements[key]

    def __setitem__(self, key, value):
        # If invoked as grid[row][val], the first grid[row] will be use __getitem__, and then the latter is a usual dictionary lookup
        if not isinstance(key, tuple):
            error('Invalid key.')
        row, col = key
        self.check_indexes(row, col);
        self._elements[row][col] = value

    def __delitem__(self, key):
        if not isinstance(key, tuple):
            error('Invalid key.')
        self.check_indexes(row, col);
        del self._elements[row][col]

    def __iter__(self):
        for row in range(self.num_rows):
            for col in range(self.num_cols):
                if (row, col) in self:
                    yield self[row, col]

    def __len__(self):
        return sum(len(row) for row in self._elements)

    def check_indexes(self, row, col):
        if not self.in_bounds(row, col):
            error('SparseGrid: ({}, {}) is outside of valid range'.format(row, col)) # TODO: better errors

    def in_bounds(self, row, col):
        return 0 <= row < self.num_rows and 0 <= col < self.num_cols

    def fill(self, value):
        # Optimization: Instead of filling the whole grid
        # (and thus populating a huge number of cells, fill by
        # clearing the existing elements and setting a new default
        # value!)
        self._elements.clear()
        # TODO(sredmond): If value is mutable, all copies will share
        # the same reference. This might be bad for students.
        # Consider forcing this to be filled with a 0-arg callable.
        self._elements.default_factory = lambda: value
        for row in range(self.num_rows):
            for col in range(self.num_cols):
                self[row, col] = value

    def resize(self, ):
        pass

    def __le__(self, other):
        rows = max(self.height, other.height)
        cols = max(self.width, other.width)


    def __eq__(self, other):
        # optimization: if literally same grid, stop
        if self is other: return True
        if self.num_rows != other.num_rows or self.num_cols != other.num_cols: return False
        for row in range(self.num_rows):
            for col in range(self.num_cols):
                if (row, col) in self:
                    # I have data there; she must, too, and it must be the same data
                    if (row, col) not in other or self[row, col] != other[row, col]:
                        return False
                else:
                    # I don't have data there; she must also not have it there
                    if (row, col) in other:
                        return False
        return True

    def __str__(self):
        pass

    def to_string_2d(self):
        pass

    def serialize(self):
        pass

    def from_stream(self):
        pass

    def randomElement(self):
        pass

    # Alternate names for cols and rows.
    @property
    def width():
        return self.cols

    @property
    def height():
        return self.rows



"""Schedule time-delayed functions into the future.

To pause current execution, use the :meth:`pause` method.

The usual warnings about using time-based functions apply in this module.
Namely, the functions here are only as precise as the underlying system time
used by the backend.

For more details, see: https://docs.python.org/3/library/time.html
"""
import campy.private.platform as _platform

import functools
import time
import logging

# Module-level logger.
logger = logging.getLogger(__name__)


def pause(milliseconds, step=None, time_fn=time.time, sleep_fn=time.sleep):
    """Pause the current thread while pumping queued events.

    This method pauses for at least at long as the requested number of
    milliseconds.

    To pause for one second::

        window = GWindow()
        oval = GOval(200, 400)
        window.add(oval, 50, 50)
        pause(1000)
        rect = GRect(200, 400)
        window.add(rect, 50, 50)

    While the execution is paused, queued events are flushed through the
    backend until the desired time has passed. This rate can be controlled with
    the step argument, which forces true thread sleep between event queue
    flushes.

    :param milliseconds: The number of milliseconds to pause for.
    :param step: The number of milliseconds to step between events.
    :param time_fn: The time function to use, measured in seconds.
    :param sleep_fn: The sleep function to use, measured in seconds.
    """
    # TODO(sredmond): Consider hiding this all in a timer_pause backend method.
    logger.info('Pausing for {} milliseconds.'.format(milliseconds))
    events = 0  # Number of events processed - useful only for debugging.
    start = time_fn()
    # Loop until at least this much time has passed.
    while time_fn() - start < milliseconds / 1000:
        _platform.Platform().event_pump_one()
        events += 1
        if step:
            sleep_fn(step / 1000)
    end = time_fn()
    logger.info('Seconds Elapsed: {}'.format(end - start))
    logger.info('Events Handled: {}'.format(events))


def repeat_every(delay_ms, function=None, fargs=()):
    """Repeat a function periodically and infinitely.

    This will schedule a delay between function calls, so if the function itself
    takes a long time, the gap between function calls may be larger than the
    supplied argument.

    This function can be used either as a decorator or directly. If used as a
    decorator, the decorated function must be invoked to start the loop. If
    used directly, the looping function will be started immediately.)

    Invoked as a decorator::

        @repeat_every(1000)
        def update_time_elapsed(label):
            label.text = 'Current Time: {}'.format(time.ctime())

        window = GWindow()
        label = GLabel()
        window.add(label, window.width / 2, window.height / 2)
        update_time_elapsed(label)  # Start the infinite loop.

    Note that the new function is explicitly invoked to start the loop.

    Invoked directly:

        def update_time_elapsed(label):
            label.text = 'Current Time: {}'.format(time.ctime())

        repeat_every(1000, update_time_elapsed, fargs=(label,))  # No additional invoking.

    :param delay_ms: The number of milliseconds to delay between invocations.

    """
    # TODO(sredmond): This is a wonky API. Consider removing or restructuring.
    if not function:  # Used as a decorator-decorator
        def decorator(fn):
            @functools.wraps(fn)
            def wrapper(*args, **kwargs):
                out = fn(*args, **kwargs)
                schedule(wrapper, delay_ms)  # Reschedule this wrapper.
                return out
            # Don't explicitly schedule the decorated function to start.
            return wrapper
        return decorator
    else:  # Invoked directly.
        def wrapper(*args, **kwargs):
            out = fn(*args, **kwargs)
            schedule(wrapper, delay_ms, fargs=fargs)  # Reschedule this wrapper.
            return out
        schedule(wrapper, delay_ms, fargs=fargs)


def schedule(function, delay_ms, fargs=()):
    """Schedule a function to execute in the future.

    Usage::

        window = GWindow()
        oval = GOval(200, 400)
        window.add(oval, 50, 50)

        def draw_box(window):
            rect = GRect(200, 400)
            window.add(rect, 50, 50)

        schedule(draw_box, 1000, fargs=(window, ))

    This instructs the backend to execute a callable after some time has passed.
    :param function: The callable to schedule.
    :param delay_ms: The number of milliseconds in the future to schedule.
    :param fargs: Additional arguments to be passed to the function.
    """
    curried = functools.partial(function, *fargs)
    # HACK(sredmond): Curried functions don't get __name__ set, but some backends rely on it.
    curried.__name__ = 'partial_{}'.format(function.__name__)
    _platform.Platform().timer_schedule(curried, delay_ms)




"""Utility functions for working with 2D geometry degrees, vectors, and constants.

Most Python libraries (such as the ``math`` module) expect angles to be specified
in radians. While this is a good convention, new students may have trouble using
radians, so this module provides degree-accepting variants of the standard trig
functions, as well as a utility to convert between radians and degrees.

Students often also represent 2D vectors by their components, so we also provide
conversions to vector magnitude and vector angle.

The mathematical constants ``PI`` and ``E`` are also exported so that students
don't have to import the separate math module for those constants.

With all of these functions, the numerical precision is only as accurate as that
provided by the underlying ``math`` module.
"""
import math

from campy.system.error import error


"""The mathematical constant `pi`, equal to the ratio of a circle's circumference
to its diameter."""
PI = math.pi


"""The mathematical constant `e`, equal to the base of the natural logarithm."""
E = math.e


def to_degrees(radians):
    """Convert an angle from radians to degrees.

    Usage::

        print(to_degrees(PI))  # => 180.0
        print(to_degrees(PI / 2))  # => 90.0

    There is no restriction of the domain of this function, but that also means
    that the codomain is unrestricted. As a concrete example::

        print(to_degrees(-PI))  # => -180.0
        print(to_degrees(4 * PI))  # => 720.0

    :param radians: An angle (in radians)
    :returns: The same angle (in degrees)
    """
    return math.degrees(radians)


def to_radians(degrees):
    """Convert an angle from degrees to radians.

    Usage::

        print(to_radians(180.0))  # => 3.141592653589793
        print(to_radians(90.0))  # => 1.5707963267948966

    There is no restriction of the domain of this function, but that also means
    that the codomain is unrestricted. As a concrete example::

        print(to_radians(-180.0))  # => -3.141592653589793
        print(to_radians(720))  # => 12.566370614359172

    :param degrees: An angle (in degrees)
    :returns: The same angle (in radians)
    """
    return math.radians(degrees)

def sin_degrees(angle):
    """Return the trigonometric sine of an angle (in degrees).

    Usage::

        print(sin_degrees(30))  # => 0.49999999999999994
        print(sin_degrees(0))  # => 0.0
        print(sin_degrees(45.0))  # => 0.7071067811865475

    :param angle: An angle (in degrees)
    :returns: The sine of the given angle
    """
    return math.sin(to_radians(angle))

def cos_degrees(angle):
    """Return the trigonometric cosine of an angle (in degrees).

    Usage::

        print(cos_degrees(60))  # => 0.5000000000000001
        print(cos_degrees(90))  # => 6.123233995736766e-17
        print(cos_degrees(45.0))  # => 0.7071067811865476

    :param angle: An angle (in degrees)
    :returns: The cosine of the given angle
    """
    return math.cos(to_radians(angle))

def tan_degrees(angle):
    """Return the trigonometric tangent of an angle (in degrees).

    Usage::

        print(tan_degrees(0))  # => 0.0
        print(tan_degrees(90))  # => 1.633123935319537e+16
        print(tan_degrees(45.0))  # => 0.9999999999999999

    :param angle: An angle (in degrees)
    :returns: The tangent of the given angle
    """
    return math.tan(to_radians(angle))


def vector_magnitude(x, y):
    """Compute the magnitude of a vector given by two components.

    To compute the magnitude of vector with two components::

        print(vector_magnitude(3, 4))  # => 5.0

    To compute the magnitude of a :class:`GPoint`::

        pt = GPoint(3, 4)
        print(vector_magnitude(*pt))  # => 5.0

    :param x: x-coordinate of vector
    :param y: y-coordinate of vector
    :returns: magnitude of the vector (x, y)
    """
    return math.hypot(x, y)


def vector_angle(x, y):
    """Return the angle (in degrees) from the origin to the given point.

    The angle is measured counterclockwise from the positive x-axis, as is standard.

    This function accounts for the fact that, in our graphical coordinate system,
    the y-axis is flipped (with respect to the traditional Cartesian plane). That
    is, in the graphical coordinate system, the y-coordinate of a point increases
    as that point descends on the screen.

    Usage::

        print(vector_angle(3, 3))  # => -45.0
        print(vector_angle(3, 3 * math.sqrt(3)))  # => -60.00000000000001
        print(vector_angle(3 * math.sqrt(3), 3))  # => -29.999999999999996

    The returned angles in these examples are negative because the y-coordinate
    represents a point in the fourth quadrant.

    :param x: x-coordinate of point
    :param y: y-coordinate of point
    :returns: angle (in degrees) from (0, 0) to (x, -y)
    """
    return to_degrees(math.atan2(-y, x))


def count_digits(n, base=10):
    """Count the number of digits in a number in some base.

    Any non-integral part of the number is discarded: ``count_digits(3.1415)``
    is the same as ``count_digits(3)``.

    The provided n must be finite and the provided base must be a positive
    integer, otherwise an error is raised.

    Usage::

        count_digits(4)  # => 1
        count_digits(41)  # => 2
        count_digits(413)  # => 3

        count_digits(9.9)  # => 1
        count_digits(10.1)  # => 2

        count_digits(-15)  # => 2
        count_digits(-314.15)  # => 3

        count_digits(3, base=2)  # => 2
        count_digits(4, base=2)  # => 3
        count_digits(8, base=3)  # => 2
        count_digits(41, base=5)  # => 3

    Invalid Usage::

        count_digits(float('inf'))  # raises error
        count_digits(5, base=-3)  # raises error
        count_digits(5, base=2.2)  # raises error

    :param n: number whose digits to count
    :param base: base system to use
    :returns: the number of digits of n when written in the given base.
    """
    if not math.isfinite(n):
        error("n must be finite")
    if base <= 0 or not float.is_integer(base):
        error("base must be a positive integer")
    n = int(abs(n))
    base = int(base)

    count = 0
    while n > 0:
        count += 1
        n //= count
    return count


# TODO(sredmond): Consider adding an isclose method.


__all__ = [
    'PI', 'E',
    'to_degrees', 'to_radians',
    'sin_degrees', 'cos_degrees', 'tan_degrees',
    'vector_magnitude', 'vector_angle',
    'count_digits'
]



#!/usr/bin/env python3 -tt
"""
This file defines the event types used in the Stanford Portable Graphics Libraries.
The structure of this package is adapted from the Java event model.

EventClassType
==============
This _enumeration definest the vent classes. The element values are each a single bit
and can be added or ORed together to generate an event mask. The CLICK_EVENT class
responds only to the MOUSE_CLICKED event type. The ANY_EVENT class selects any event.

    - NULL_EVENT
    - ACTION_EVENT
    - KEY_EVENT
    - TIMER_EVENT
    - WINDOW_EVENT
    - MOUSE_EVENT
    - CLICK_EVENT
    - ANY_EVENT

EventType
=========
This _enumeration type defines the event types for all events.

    - WINDOW_CLOSED
    - WINDOW_RESIZED
    - ACTION_PERFORMED
    - MOUSE_CLICKED
    - MOUSE_PRESSED
    - MOUSE_RELEASED
    - MOUSE_MOVED
    - MOUSE_DRAGGED
    - KEY_PRESSED
    - KEY_RELEASED
    - KEY_TYPED
    - TIMER_TICKED


ModifierCodes
=============
This _enumeration type defines a set of constants used to check whether
modifiers are in effect.

    - SHIFT_DOWN
    - CTRL_DOWN
    - META_DOWN
    - ALT_DOWN
    - ALT_GRAPH_DOWN
    - BUTTON1_DOWN
    - BUTTON2_DOWN
    - BUTTON3_DOWN

KeyCodes
========
This type defines the names of the key codes returned in a key event

    - BACKSPACE_KEY
    - TAB_KEY
    - ENTER_KEY
    - CLEAR_KEY
    - ESCAPE_KEY
    - PAGE_UP_KEY
    - PAGE_DOWN_KEY
    - END_KEY
    - HOME_KEY
    - LEFT_ARROW_KEY
    - UP_ARROW_KEY
    - RIGHT_ARROW_KEY
    - DOWN_ARROW_KEY
    - F1_KEY
    - F2_KEY
    - F3_KEY
    - F4_KEY
    - F5_KEY
    - F6_KEY
    - F7_KEY
    - F8_KEY
    - F9_KEY
    - F10_KEY
    - F11_KEY
    - F12_KEY
    - DELETE_KEY
    - HELP_KEY

TODO:
    Reconsider using EventClassType. Instead use inheritance?
    Deduplicate events that capture a GWindow.
    Check for validity on property access.
"""
import enum as _enum


@_enum.unique
class EventClassType(_enum.Enum):
    NULL_EVENT   = 0x000
    ACTION_EVENT = 0x010
    KEY_EVENT    = 0x020
    TIMER_EVENT  = 0x040
    WINDOW_EVENT = 0x080
    MOUSE_EVENT  = 0x100
    CLICK_EVENT  = 0x200
    ANY_EVENT    = 0x3F0


@_enum.unique
class EventType(_enum.Enum):
    WINDOW_CLOSED    = EventClassType.WINDOW_EVENT.value + 1
    WINDOW_RESIZED   = EventClassType.WINDOW_EVENT.value + 2
    CONSOLE_CLOSED   = EventClassType.WINDOW_EVENT.value + 3
    ACTION_PERFORMED = EventClassType.ACTION_EVENT.value + 1
    MOUSE_CLICKED    = EventClassType.MOUSE_EVENT.value + 1
    MOUSE_PRESSED    = EventClassType.MOUSE_EVENT.value + 2
    MOUSE_RELEASED   = EventClassType.MOUSE_EVENT.value + 3
    MOUSE_MOVED      = EventClassType.MOUSE_EVENT.value + 4
    MOUSE_DRAGGED    = EventClassType.MOUSE_EVENT.value + 5
    KEY_PRESSED      = EventClassType.KEY_EVENT.value + 1
    KEY_RELEASED     = EventClassType.KEY_EVENT.value + 2
    KEY_TYPED        = EventClassType.KEY_EVENT.value + 3
    TIMER_TICKED     = EventClassType.TIMER_EVENT.value + 1


@_enum.unique
class ModifierCodes(_enum.Enum):
    SHIFT_DOWN     = 1 << 0
    CTRL_DOWN      = 1 << 1
    META_DOWN      = 1 << 2
    ALT_DOWN       = 1 << 3
    ALT_GRAPH_DOWN = 1 << 4
    BUTTON1_DOWN   = 1 << 5
    BUTTON2_DOWN   = 1 << 6
    BUTTON3_DOWN   = 1 << 7


@_enum.unique
class KeyCodes(_enum.Enum):
    BACKSPACE_KEY = 8
    TAB_KEY = 9
    ENTER_KEY = 10
    CLEAR_KEY = 12
    ESCAPE_KEY = 27
    PAGE_UP_KEY = 33
    PAGE_DOWN_KEY = 34
    END_KEY = 35
    HOME_KEY = 36
    LEFT_ARROW_KEY = 37
    UP_ARROW_KEY = 38
    RIGHT_ARROW_KEY = 39
    DOWN_ARROW_KEY = 40
    F1_KEY = 112
    F2_KEY = 113
    F3_KEY = 114
    F4_KEY = 115
    F5_KEY = 116
    F6_KEY = 117
    F7_KEY = 118
    F8_KEY = 119
    F9_KEY = 120
    F10_KEY = 121
    F11_KEY = 122
    F12_KEY = 123
    DELETE_KEY = 127
    HELP_KEY = 156

class GEvent:
    """
    This class is the root of the hierarchy for all events.

    The standard paradigm for using GEvent is illustrated
    by the following program, which allows the user to draw lines on the
    graphics window::

        gw = _gwindow.GWindow()
        print("This program lets the user draw lines by dragging.")
        while(True):
            e = gevents.waitForEvent(gevents.EventClassType.MOUSE_EVENT)
            if(e.getEventType() == gevents.EventType.MOUSE_PRESSED):
                line = gobjects.GLine(e.getX(), e.getY(), e.getX(), e.getY());
                gw.add(line);
            elif (e.getEventType() == gevents.EventType.MOUSE_DRAGGED)
                line.setEndPoint(e.getrX(), e.getY());

    Attributes:
        event_class [EventClassType]: Enumerated type constant indicating the class of the event.
        event_type [EventType]: Enumerated type constant corresponding to the specific event type.
        time [float]: System time in milliseconds at which the event occurred.
        valid [bool]: Whether this event represents a fully-initialized valid event.
        modifiers [int]: Integer whose bits indicate what modifiers are in effect.

    Notes:
        To ensure portability among systems that represent time in different
        ways, this library uses a float to
        represent time, which is always encoded as the number of milliseconds
        that have elapsed since 00:00:00 UTC on January 1, 1970, which is
        the conventional zero point for computer-based time systems.

        To check whether the shift key is down, for example, one could use
        the following code::

            if e.modifiers & SHIFT_DOWN: ...
    """

    def __init__(self):
        '''
        Ensures that an event is properly initialized to a NULL event.

        @rtype: void
        '''
        self._event_class = EventClassType.NULL_EVENT
        self._event_type = None
        self._valid = False

        self._time = None
        self._modifiers = 0

    # "Enforce" read-only properties.
    # TODO(sredmond): This is somewhat gross and ill-advised.

    @property
    def event_class(self):
        return self._event_class

    @property
    def event_type(self):
        return self._event_type

    @property
    def valid(self):
        return self._valid

    @property
    def time(self):
        return self._time

    @property
    def modifiers(self):
        return self._modifiers

    def __str__(self):
        return "GEvent(NULL)";

class GWindowEvent(GEvent):
    '''
    This event subclass represents a window event.
    Each GWindowEvent keeps track of the event type
    (WINDOW_CLOSED, WINDOW_RESIZED) along
    with the identity of the window.

    Attributes:
        gwindow [GWindow]: Reference to the GWindow in which this event took place
    '''

    def __init__(self, event_type=None, gwindow=None):
        '''
        Creates a GWindowEvent using the specified parameters

        @type type: EventType
        @param type: type of event
        @type gw: GWindow
        @param gw: GWindow event took place in
        '''
        super().__init__()
        if event_type != None and gwindow != None:
            self._event_class = EventClassType.WINDOW_EVENT
            self._event_type = event_type
            self._valid = True

        self._gwindow = gwindow

    @property
    def gwindow(self):
        # TODO(sredmond): Do we have to recreate a new GWindow with the existing data?
        return self._gwindow

    # def getGWindow(self):
    #     '''
    #     Returns the graphics window in which this event occurred.

    #     @rtype: L{GWindow}
    #     '''
    #     import campy.graphics.gwindow as _gwindow
    #     return _gwindow.GWindow(gwd = self.gwd)

    def __str__(self):
        if not self.valid:
            return "GWindowEvent(?)"
        return  "GWindowEvent({})".format(self.event_type.name)

class GActionEvent(GEvent):
    '''
    This event subclass represents an action event.
    Action events are generated by the classes in the
    L{GInteractor}
    hierarchy.  As an example, the following program displays
    a button that, when pushed, generates the message
    &ldquo;Please do not press this button again&rdquo;
    (with thanks to Douglas Adams&rsquo;s Hitchhiker&rsquo;s
    Guide to the Galaxy)::

        gw = _gwindow.GWindow
        button = ginteractors.GButton("RED");
        gw.addToRegion(button, "SOUTH");
        while(True):
            e = gevents.waitForEvent(ACTION_EVENT | CLICK_EVENT);
            if(e.getEventType() == MOUSE_CLICKED):
                break;
            print("Please do not press this button again.")

    Attributes:
        source [GObject]: GInteractor from which event originated.
        action_command [string]: GInteractor command.
    '''

    def __init__(self, event_type=None, source=None, action_command=None):
        '''
        Creates a GActionEvent using the specified parameters.

        @type type: EventType
        @param type: type of event
        @type source: GObject
        @param source: interactor event originated from
        @type actionCommand: string
        @param actionCommand: interactor command
        @rtype: void
        '''
        super().__init__()
        if event_type != None and source != None and action_command != None:
            self._event_class = EventClassType.ACTION_EVENT
            self._event_type = event_type
            self._valid = True

        self._source = source
        self._action_command = action_command

    @property
    def source(self):
        return self._source

    @property
    def action_command(self):
        return self._action_command

    def __str__(self):
        if not valid:
            return "GActionEvent(?)"
        return "GActionEvent({}, {})".format(self.event_type.name, self.action_command)

class GMouseEvent(GEvent):
    '''
    This event subclass represents a mouse event.  Each mouse event
    records the event type (MOUSE_PRESSED,
    MOUSE_RELEASED, MOUSE_CLICKED,
    MOUSE_MOVED, MOUSE_DRAGGED) along
    with the coordinates of the event.  Clicking the mouse generates
    three events in the following order: MOUSE_PRESSED,
    MOUSE_RELEASED, MOUSE_CLICKED.

    As an example, the following program uses mouse events to let
    the user draw rectangles on the graphics window.  The only
    complexity in this code is the use of the library functions
    min and abs to ensure that the
    dimensions of the rectangle are positive::

        gw = _gwindow.GWindow
        print("This program lets the user draw rectangles.")
        while(True):
            e = gevents.waitForEvent();
            if (e.getEventType() == MOUSE_PRESSED):
                startX = e.getX();
                startY = e.getY();
                rect = gobjects.GRect(startX, startY, 0, 0);
                rect.setFilled(True);
                gw.add(rect);
            elif(e.getEventType() == MOUSE_DRAGGED):
                x = min(e.getX(), startX);
                y = min(e.getY(), startY);
                width = abs(e.getX() - startX);
                height = abs(e.getY() - startY);
                rect.setBounds(x, y, width, height);

    Attributes:
        gwindow [GWindow]: GWindow in which MouseEvent occurred.
        x [float]: The x-coordinate at which the event occurred.
        y [float]: The y-coordinate at which the event occurred.

    Notes:
        The x- and y-coordinates are given relative to the window origin at
        the upper left corner of the window.
    '''

    def __init__(self, event_type=None, gwindow=None, x=None, y=None):
        '''
        Creates a GMouseEvent using the specified parameters.

        @type type: EventType
        @param type: type of event
        @type gw: GWindow
        @param gw: window event took place in
        @type x: float
        @param x: x location of event
        @type y: float
        @param y: y location of event
        @rtype: void
        '''
        super().__init__()
        if type != None and gwindow != None and x != None and y != None:
            self._event_class = EventClassType.MOUSE_EVENT
            self._event_type = event_type
            self._valid = True

        self._gwindow = gwindow
        self._x = x
        self._y = y

    @property
    def gwindow(self):
        # TODO(sredmond): Do we have to recreate a new GWindow with the existing data?
        return self._gwindow

    @property
    def x(self):
        return self._x

    @property
    def y(self):
        return self._y

    def __str__(self):
        if not valid:
            return "GMouseEvent(?)"
        return "GMouseEvent({}, x={}, y={})".format(self.event_type.name, self.x, self.y)

class GKeyEvent(GEvent):
    '''
    This event subclass represents a key event.  Each key event records
    the event type along with two representations of the key.  The
    getKeyChar function is more generally useful and
    returns the character after taking into account modifier keys.
    The getKeyCode function returns an integer identifying
    the key, which can be a function key as well as a standard key.
    The codes return by getKeyCode are listed in the
    KeyCodes _enumeration.

    Attributes:
        gwindow [GWindow]: GWindow in which MouseEvent occurred.
        key_char [int]: Value of key.
        key_code [int]: Code of key.

    Notes:
        Returns the character represented by the keystroke, taking the modifier
        keys into account.  For example, if the user types the 'a'
        key with the shift key down, getKeyChar will return
        'A'.  If the key code in the event does not correspond
        to a character, getKeyChar returns the null character.
    '''

    def __init__(self, event_type=None, gwindow=None, key_char=None, key_code=None):
        '''
        Creates a GKeyEvent using the specified parameters.

        @type type: EventType
        @param type: type of event
        @type gw: GWindow
        @param gw: window event takes place in
        @type keyChar: int
        @param keyChar: value of key
        @type keyCode: KeyCodes
        @param keyCode: code of key
        @rtype: void
        '''
        super().__init__()
        if event_type != None and gwindow != None and key_char != None and key_code != None:
            self._event_class = EventClassType.KEY_EVENT
            self._event_type = event_type
            self._valid = True

        self._gwindow = gwindow
        self._key_char = key_char
        self._key_code = key_code

    @property
    def gwindow(self):
        # TODO(sredmond): Do we have to recreate a new GWindow with the existing data?
        return self._gwindow

    @property
    def key_char(self):
        return self._key_char

    @property
    def key_code(self):
        return self._key_code

    def __str__(self):
        if not valid:
            return "GKeyEvent(?)"
        # TODO(sredmond): The C++ libs use key_char for KEY_TYPED for some reason.
        # Investigate further.
        return "GKeyEvent({}, {})".format(self.event_type.name, chr(self.key_code))

class GTimerEvent(GEvent):
    '''
    This event subclass represents a timer event.  Timer events are
    generated by a GTimer
    object, which produces a new event at a fixed interval measured in
    milliseconds.  As an example, the following program generates a
    timer event every two seconds, stopping when the user clicks
    somewhere in the window::


        print("This program generates timer events.")
        timer = gtimer.GTimer(2000)
        timer.start()
        while(True):
            e = gevents.waitForEvent(CLICK_EVENT | TIMER_EVENT);
            if (e.getEventType() == MOUSE_CLICKED):
                break;
        print("Timer ticked")

    Attributes:
        timer [GTimer]: GTimer from which this event originated.
    '''

    def __init__(self, event_type=None, timer=None):
        '''
        Creates a GTimerEvent for the specified timer.

        @type type: EventType
        @param type: the event type
        @type timer: GTimer
        @param timer: timer associated with event
        @rtype: void
        '''
        super().__init__()
        if event_type != None and timer != None:
            self._event_class = EventClassType.TIMER_EVENT
            self._event_type = event_type
            self._valid = True

        self._timer = timer

    @property
    def timer(self):
        return self._timer

    def __str__():
        '''
        Converts the event to a human-readable representation of the event.

        @rtype: string
        '''
        if not valid:
            return "GTimerEvent(?)"
        return "GTimerEvent({})".format(self.event_type.name)

def wait_for_click():
    '''
    Waits for a mouse click in any window, discarding any other events.

    @rtype: void
    '''
    wait_for_event(EventClassType.CLICK_EVENT)

def wait_for_event(mask=EventClassType.ANY_EVENT):
    '''
    Dismisses the process until an event occurs whose type is covered by
    the event mask.  The mask parameter is a combination of the events of
    interest.  For example, to wait for a mouse event or an action event,
    clients can use the following call:

    e = waitForEvent(MOUSE_EVENT + ACTION_EVENT);

    The mask parameter is optional.  If it is missing,
    waitForEvent accepts any event.

    As a more sophisticated example, the following code is the canonical
    event loop for an animated application that needs to respond to mouse,
    key, and timer events::

        timer = gtimer.GTimer(ANIMATION_DELAY_IN_MILLISECONDS);
        timer.start()
        while(True):
            e = gevents.waitForEvent(TIMER_EVENT + MOUSE_EVENT + KEY_EVENT)
            if(e.getEventClass == gevents.EventClassType.TIMER_EVENT):
                takeAnimationStep()
            elif(e.getEventClass == gevents.EventClassType.MOUSE_EVENT):
                handleMouseEvent(e)
            elif(e.getEventClass == gevents.EventClassType.KEY_EVENT):
                handleKeyEvent(e)

    @type mask: EventClassType
    @param mask: EventClassType to wait for, defaults to ANY_EVENT
    @rtype: GEvent
    '''
    import campy.private.platform as _platform
    return _platform.Platform().waitForEvent(mask)

def get_next_event(mask=EventClassType.ANY_EVENT):
    '''
    Checks to see if there are any events of the desired type waiting on the
    event queue.  If so, this function returns the event in exactly the same
    fashion as waitForEvent; if not, getNextEvent
    returns an invalid event.  The mask parameter is optional.
    If it is missing, getNextEvent accepts any event.

    @type mask: EventClassType
    @param mask: EventClasstype to check for, defaults to ANY_EVENT
    @rtype: GEvent
    '''
    import campy.private.platform as _platform
    return _platform.Platform().getNextEvent(mask)



def error(str):
    print(str)


class Graph():
    def __init__(self, NodeType, ArcType, src=None):
        self._NodeType = NodeType
        self._ArcType = ArcType
        self.nodes = set()
        self.arcs = set()
        self.node_map = dict()
        # self.comparator = None  # TODO
        if src:
            self._deep_copy(src)

    def add_arc_by_nodes(self, n1, n2, named=False):
        if named:
            n1 = self._get_node_by_name(n1, member='add_arc')
            n2 = self._get_node_by_name(n1, member='add_arc')
        self._verify_existing_node(n1)
        self._verify_existing_node(n2)
        arc = self._ArcType()
        arc.start = n1
        arc.finish = n2

    def add_arc(self, arc):
        self._verify_not_none(arc)
        if not self._node_exists(arc.start):
            self.add_node(arc.start)
        if not self._node_exists(arc.end):
            self.add_node(arc.end)
        arc.start.arcs.add(arc)
        self.arcs.add(arc)
        return arc

    def add_node(self, node_info, named=False):
        if named:
            node = self._get_node_by_name(node_info, member='add_node')
            if not node:
                node = self._NodeType()
                node.arcs = set()
                node.name = node_info
        else:
            node = node_info
        self._verify_not_none(node, 'add_node')
        if node.name in self.node_map:
            error("Graph::add_node: node {} already exists".format(node.name))

        self.nodes.add(node)
        self.node_map[node.name] = node
        return node

    def clear(self):
        self.nodes = set()
        self.arcs = set()
        self.node_map = dict()

    def get_arc_set(self, node_info, named=False):
        if named:
            node = self._get_node_by_name(node_info)
        else:
            node = node_info
        self._verify_existing_node(node, member='get_arc_set')
        return node.arcs

    def get_neighbors(self, node_info, named=False):
        if named:
            node = self._get_node_by_name(node_info)
        else:
            node = node_info
        self._verify_existing_node(node, member='get_neighbors')
        return {arc.finish for arc in node.arcs}

    def is_connected(self, n1, n2, named=False):
        if named:
            n1 = self._get_node_by_name(n1, member='add_arc')
            n2 = self._get_node_by_name(n1, member='add_arc')
        # TODO error checking here
        for arc in n1.arcs:
            if arc.finish == n2:
                return True
        return False

    def is_empty(self):
        return len(self.nodes) == 0

    def remove_arc_by_nodes(self, n1, n2, named=False):
        if named:
            n1 = self._get_node_by_name(n1)
            n2 = self._get_node_by_name(n1)
            return self.remove_arc_by_nodes(n1, n2, named=False)
        if not self._node_exists(n1) or not self._node_exists(n2):
            return
        to_remove = []
        for arc in self.arcs:
            if arc.start == n1 and arc.finish == n2:
                to_remove.append(arc)

        for arc in to_remove:
            self.remove_arc(arc)

    def remove_arc(self, arc):
        arc.start.arcs.discard(arc)
        self.arcs.discard(arc)

    def remove_node(self, node, named=False):
        if named:
            node = self._get_node_by_name(node)
            self.remove_node(node, named=True)
        if not self._node_exists(node):
            return

        to_remove = []
        for arc in self.arcs:
            if arc.start == node or arc.end == node:
                to_remove.append(arc)
        for arc in to_remove:
            self.remove_arc(arc)

        self.nodes.discard(node)
        del self.node_map[node.name]

    def scan_arc_data(self):
        pass

    def scan_node_data(self):
        pass

    def scan_graph_entry(self, stream):
        pass

    def write_arc_data(self, ostream):
        pass

    def write_node_data(self, ostream):
        pass

    @property
    def size(self):
        pass

    # Implementation details, not useful to client
    def _deep_copy(self):
        pass

    def _get_node_by_name(self, name, member=''):
        node = self.node_map.get(name)
        if not node:
            error('Graph::{}: no node named {}'.format(member, name))
        return node

    def _graph_compare(self, other):
        pass

    def _arc_exists(self, arc):
        return arc in self.arcs

    def _node_exists(self, node):
        return node and self.node_map.get(node.name) == node

    def _verify_existing_node(self, node, member=''):
        self._verify_not_none(node, member)
        if not self._node_exists(node):
            error('Graph::{}: node not found in graph'.format(member))

    def _verify_not_none(self, value, member=''):
        if value is None:
            error('Graph::{}: parameter cannot be None'.format(member))

    def _scan_node(self, scanner):
        pass

    def __repr__(self):
        return "Graph"  # TODO obviously

    def __eq__(self, other):
        return False  # TODO, obviously

    # Iterating?



# TODO(sredmond): Possibly rename this file priorityqueue.h for compatibility.
"""
Min-heap.
"""
import collections as _collections

# TODO(sredmond): Make this into a collection subclass.
# TODO(sredmond): Use namedtuples.
import heapq as _heapq

class PriorityQueue(object):
    def __init__(self, initializer=None, priorities_first=True):
        if not initializer:
            initializer = []

        if priorities_first:
            self._heap = [(priority, value) for priority, value in initializer]
        else:
            self._heap = [(priority, value) for value, priority in initializer]
        _heapq.heapify(self._heap)

    def add(self, value, priority):
        _heapq.heappush(self._heap, (priority, value))

    def back(self):
        raise NotImplementedError

    def change_priority(self, value, new_priority):
        raise NotImplementerError

    def clear(self):
        self._heap = []

    def remove(self):
        top = _heapq.heappop(self._heap)
        return top[1]

    def front(self):
        raise NotImplementedError

    def __len__(self):
        return len(self._heap)

    def __eq__(self, other):
        return self._heap == other._heap

    def peek(self):
        top = self._heap[0]
        return top

    def peek_value(self):
        top = self._heap[0]
        return top[1]

    def peek_priority(self):
        top = self._heap[0]
        return top[0]

    enqueue = add
    dequeue = remove

    def __str__(self):
        pass




